from human_cver.datasets.Pose2Mesh.Human36M.dataset import Human36M
from human_cver.models.smpl.smpl import SMPL
from human_cver.vis.dataset_viewer import show_dataset

smpl = SMPL(model_path="/home/lwk/human_cver/data", joint_type="SMPL")
show_dataset(Human36M(smpl=smpl))



import os
import os.path as osp

import numpy as np
import torch
import torch.nn as nn
from smplx import SMPL as _SMPL
from smplx.lbs import vertices2joints

from .kp_map import JOINT_MAP, JOINT_NAMES, OPENPOSE_JOINT_NAMES, SMPL_JOINT_NAMES

J_regressor_extra_fname = osp.join(osp.dirname(__file__), "J_regressor_extra.npy")
J_regressor_h36m_correct_fname = osp.join(osp.dirname(__file__), "J_regressor_h36m_correct.npy")
J_regressor_coco_fname = osp.join(osp.dirname(__file__), "J_regressor_coco.npy")  # copy from Pose2Mesh


class SMPLBase(_SMPL):
    def __init__(self, model_path: str, gender: str = "neutral"):
        super().__init__(
            model_path=model_path,
            gender=gender,
            create_betas=False,
            create_body_pose=False,
            create_global_orient=False,
            create_transl=False,
            use_hands=True,
            use_feet_keypoints=True,
        )

def get_gender(gender: str):
    d = {
        "f": "female",
        "female": "female",
        "m": "male",
        "male": "male",
        "n": "neutral",
        "neutral": "neutral",
    }
    gender = gender.lower()
    if gender in d:
        return d[gender]
    assert 0, f"no gender: {gender}"
    return None

class SMPL(nn.Module):
    r"""Extension of the official SMPL implementation to support more joints

    Args:
        model_path:
            SMPL_FEMALE.pkl
            SMPL_MALE.pkl
            SMPL_NEUTRAL.pkl
        gender: str: neutral, male, female
        joint_type: str: J49(OpenPose+SMPL), SMPL(24), OpenPose(25)
    """

    def __init__(self, model_path: str, gender: str = "neutral", joint_type="J49"):
        super().__init__()
        self.smpl_dict = {
            "female": SMPLBase(model_path, gender="female"),
            "male": SMPLBase(model_path, gender="male"),
            "neutral": SMPLBase(model_path, gender="neutral"),
        }
        self.gender = gender
        self.faces = self.smpl_dict["neutral"].faces

        joint_type = joint_type.lower()
        d = {
            "j49": JOINT_NAMES,
            "openpose": OPENPOSE_JOINT_NAMES,
            "smpl": SMPL_JOINT_NAMES,
        }
        if joint_type not in d:
            assert 0, f"no joint type:{joint_type}"

        self.joint_name = d[joint_type]

        joints = [JOINT_MAP[i] for i in self.joint_name]
        self.joint_map = torch.tensor(joints, dtype=torch.long)

        J_regressor_extra = np.load(J_regressor_extra_fname)
        assert J_regressor_extra.shape == (9, 6890)
        self.register_buffer("J_regressor_extra", torch.tensor(J_regressor_extra, dtype=torch.float32))
        self.register_buffer(
            "J_regressor_h36m_correct", torch.tensor(np.load(J_regressor_h36m_correct_fname), dtype=torch.float32)
        )
        self.register_buffer("J_regressor_coco", torch.tensor(np.load(J_regressor_coco_fname), dtype=torch.float32))

    def kps_num(self) -> int:
        return len(self.joint_name)

    def forward(self, betas, pose, gender: str = None):
        r"""
        Args:
            betas: shape: N,10
            pose: shape N,72  or N,24,3 or N,24,3,3

        Return:
            vertices: shape: N,6890,3
            joints: shape: N,49,3
        """
        bs = betas.shape[0]
        if pose.shape == (bs, 72):
            pose = pose.reshape([bs, 24, 3])

        gender = get_gender(gender) or self.gender
        smpl_output = self.smpl_dict[gender].forward(
            betas=betas,
            body_pose=pose[:, 1:],
            global_orient=pose[:, :1],
            pose2rot=len(pose.shape) == 3,
        )
        joints = smpl_output.joints
        vertices = smpl_output.vertices
        extra_joints = vertices2joints(self.J_regressor_extra, vertices)
        joints = torch.cat([joints, extra_joints], dim=1)
        joints = joints[:, self.joint_map, :]
        return vertices, joints

    def get_h36m_joints(self, vertices):
        """
        This method is used to get the joint locations from the SMPL mesh
        Input:
            vertices: size = (B, 6890, 3)
        Output:
            3D joints: size = (B, 17, 3)
        """
        joints = torch.einsum("bik,ji->bjk", [vertices, self.J_regressor_h36m_correct])
        return joints

    def get_coco_joints(self, vertices):
        """
        Return:
            3D joints: size = (B, 17, 3)
        """
        joints = torch.einsum("bik,ji->bjk", [vertices, self.J_regressor_coco])
        return joints

    def tpose(self, gender: str = None):
        r"""TPose

        Return:
            vertices: 6890,3
            joints: 49,3

        Example::
            >>> from human_cver.models.smpl.smpl import SMPL
            >>> from human_cver.io.obj_file import save_obj
            >>> smpl = SMPL(model_path="data")
            >>> vertices, joints = smpl.tpose()
            >>> h36m = smpl.get_h36m_joints(vertices[None])[0]
            >>> save_obj("tpose.obj", vertices, smpl.faces)
            >>> save_obj("joints.obj", joints)
            >>> save_obj("h36m.obj", h36m)
        """
        gender = get_gender(gender) or self.gender
        smpl_output = self.smpl_dict[gender].forward(
            betas=torch.zeros([1, 10]),
            body_pose=torch.zeros([1, 23, 3]),
            global_orient=torch.zeros([1, 1, 3]),
            pose2rot=True,
        )
        joints = smpl_output.joints
        vertices = smpl_output.vertices
        extra_joints = vertices2joints(self.J_regressor_extra, vertices)
        joints = torch.cat([joints, extra_joints], dim=1)
        joints = joints[:, self.joint_map, :]
        return vertices[0], joints[0]



# -*- coding: utf-8 -*-

# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de

# copy from: https://github.com/mkocabas/PARE/blob/master/pare/core/constants.py

import numpy as np

# Mean and standard deviation for normalizing input image
IMG_NORM_MEAN = [0.485, 0.456, 0.406]
IMG_NORM_STD = [0.229, 0.224, 0.225]

"""
We create a superset of joints containing the OpenPose joints together with the ones that each dataset provides.
We keep a superset of 24 joints such that we include all joints from every dataset.
If a dataset doesn't provide annotations for a specific joint, we simply ignore it.
The joints used here are the following:
"""
OPENPOSE_JOINT_NAMES = [
# JOINT_NAMES = [
    # 25 OpenPose joints (in the order provided by OpenPose)
    "OP Nose",
    "OP Neck",
    "OP RShoulder",
    "OP RElbow",
    "OP RWrist",
    "OP LShoulder",
    "OP LElbow",
    "OP LWrist",
    "OP MidHip",
    "OP RHip",
    "OP RKnee",
    "OP RAnkle",
    "OP LHip",
    "OP LKnee",
    "OP LAnkle",
    "OP REye",
    "OP LEye",
    "OP REar",
    "OP LEar",
    "OP LBigToe",
    "OP LSmallToe",
    "OP LHeel",
    "OP RBigToe",
    "OP RSmallToe",
    "OP RHeel",
]
SMPL_JOINT_NAMES = [
    # 24 Ground Truth joints (superset of joints from different datasets)
    "Right Ankle",
    "Right Knee",
    "Right Hip",
    "Left Hip",
    "Left Knee",
    "Left Ankle",
    "Right Wrist",
    "Right Elbow",
    "Right Shoulder",
    "Left Shoulder",
    "Left Elbow",
    "Left Wrist",
    "Neck (LSP)",
    "Top of Head (LSP)",
    "Pelvis (MPII)",
    "Thorax (MPII)",
    "Spine (H36M)",
    "Jaw (H36M)",
    "Head (H36M)",
    "Nose",
    "Left Eye",
    "Right Eye",
    "Left Ear",
    "Right Ear",
]
SMPL_ROOT_IDX = SMPL_JOINT_NAMES.index("Pelvis (MPII)")
JOINT_NAMES = OPENPOSE_JOINT_NAMES + SMPL_JOINT_NAMES
# Dict containing the joints in numerical order
JOINT_IDS = {JOINT_NAMES[i]: i for i in range(len(JOINT_NAMES))}

# Map joints to SMPL joints
JOINT_MAP = {
    "OP Nose": 24,
    "OP Neck": 12,
    "OP RShoulder": 17,
    "OP RElbow": 19,
    "OP RWrist": 21,
    "OP LShoulder": 16,
    "OP LElbow": 18,
    "OP LWrist": 20,
    "OP MidHip": 0,
    "OP RHip": 2,
    "OP RKnee": 5,
    "OP RAnkle": 8,
    "OP LHip": 1,
    "OP LKnee": 4,
    "OP LAnkle": 7,
    "OP REye": 25,
    "OP LEye": 26,
    "OP REar": 27,
    "OP LEar": 28,
    "OP LBigToe": 29,
    "OP LSmallToe": 30,
    "OP LHeel": 31,
    "OP RBigToe": 32,
    "OP RSmallToe": 33,
    "OP RHeel": 34,
    "Right Ankle": 8,
    "Right Knee": 5,
    "Right Hip": 45,
    "Left Hip": 46,
    "Left Knee": 4,
    "Left Ankle": 7,
    "Right Wrist": 21,
    "Right Elbow": 19,
    "Right Shoulder": 17,
    "Left Shoulder": 16,
    "Left Elbow": 18,
    "Left Wrist": 20,
    "Neck (LSP)": 47,
    "Top of Head (LSP)": 48,
    "Pelvis (MPII)": 49,
    "Thorax (MPII)": 50,
    "Spine (H36M)": 51,
    "Jaw (H36M)": 52,
    "Head (H36M)": 53,
    "Nose": 24,
    "Left Eye": 26,
    "Right Eye": 25,
    "Left Ear": 28,
    "Right Ear": 27,
}

ROOT_JOINT_NAME = "Pelvis (MPII)"
# Joint selectors
# Indices to get the 14 LSP joints from the 17 H36M joints
H36M_TO_J17 = [6, 5, 4, 1, 2, 3, 16, 15, 14, 11, 12, 13, 8, 10, 0, 7, 9]
H36M_TO_J14 = H36M_TO_J17[:14]
# Indices to get the 14 LSP joints from the ground truth joints
J24_TO_J17 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 18, 14, 16, 17]
J24_TO_J14 = J24_TO_J17[:14]

# Permutation of SMPL pose parameters when flipping the shape
SMPL_JOINTS_FLIP_PERM = [0, 2, 1, 3, 5, 4, 6, 8, 7, 9, 11, 10, 12, 14, 13, 15, 17, 16, 19, 18, 21, 20, 23, 22]
SMPL_POSE_FLIP_PERM = []
for i in SMPL_JOINTS_FLIP_PERM:
    SMPL_POSE_FLIP_PERM.append(3 * i)
    SMPL_POSE_FLIP_PERM.append(3 * i + 1)
    SMPL_POSE_FLIP_PERM.append(3 * i + 2)
# Permutation indices for the 24 ground truth joints
J24_FLIP_PERM = [5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 12, 13, 14, 15, 16, 17, 18, 19, 21, 20, 23, 22]
# Permutation indices for the full set of 49 joints
J49_FLIP_PERM = [0, 1, 5, 6, 7, 2, 3, 4, 8, 12, 13, 14, 9, 10, 11, 16, 15, 18, 17, 22, 23, 24, 19, 20, 21] + [
    25 + i for i in J24_FLIP_PERM
]

SMPLH_TO_SMPL = (
    np.arange(0, 156)
    .reshape((-1, 3))[np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 37])]
    .reshape(-1)
)



import glob
import os.path as osp
from typing import List

import cv2

from ..tools.logger import Logger
from .base import BaseDataset


class ImageFolderDataset(BaseDataset):
    """图片文件夹数据集"""

    def __init__(self, folder: str, fmts: List[str] = ["*.jpg", "*.jpeg", "*.png"]) -> None:
        super().__init__()

        if not osp.exists(folder):
            Logger.error(f"not exists: {folder}")
            return

        self.imgnames = []
        for fmt in fmts:
            self.imgnames.extend(glob.glob(osp.join(folder, fmt)))
        self.imgnames.sort()

    def info(self):
        return dict()

    def keys(self):
        return ["imgname", "img"]

    def __len__(self):
        return len(self.imgnames)

    def __getitem__(self, index):
        imgname = self.imgnames[index]
        img = cv2.imread(imgname)
        return {
            "imgname": imgname,
            "img": img,
        }



import os.path as osp
from typing import Dict, List

import cv2
import numpy as np

from ..tools.logger import Logger
from .base import BaseDataset


class VideoDataset(BaseDataset):
    """视频数据集"""

    def __init__(self, video_filename: str) -> None:
        super().__init__()
        self.video_filename = video_filename
        if not osp.exists(video_filename):
            Logger.error(f"not exists: {video_filename}")
            return

        self.cap = cv2.VideoCapture(video_filename)
        self.frame_width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.frame_height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)
        self._curr_frame = -1

    def info(self):
        return {
            "frame_width": self.frame_width,
            "frame_height": self.frame_height,
            "frame_count": self.frame_count,
            "fps": self.fps,
            "fname": self.video_filename,
        }

    def keys(self) -> List[str]:
        return ["img"]

    def __len__(self):
        return self.frame_count

    def __getitem__(self, index) -> Dict[str, np.ndarray]:
        if not (0 <= index < len(self)):
            Logger.error(f"bad index: {index}")
            return
        if self._curr_frame + 1 != index:
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, index)
        opened, frame = self.cap.read()
        self._curr_frame = index
        if not opened:
            Logger.error(f"can not open: {self.video_filename}")
            return

        return {
            "img": frame,
        }

    def close(self):
        self.cap.close()



import importlib
import os.path as osp
import shutil

from omegaconf import OmegaConf

from ..tools.logger import Logger


def get_obj_from_str(string: str, reload: bool = False):
    """从字符串获取实例

    Args:
        string (str): 实例字符串
        reload (bool, optional): Defaults to False.

    Returns:
        _type_: 实例
    """
    module, cls = string.rsplit(".", 1)
    if reload:
        module_imp = importlib.import_module(module)
        importlib.reload(module_imp)
    return getattr(importlib.import_module(module, package=None), cls)


def instantiate_from_config(config: dict):
    """从配置获取实例

    Args:
        config (dict): config

    Raises:
        KeyError: 必须包含target

    Returns:
        _type_: 实例
    """
    if not "target" in config:
        raise KeyError("Expected key `target` to instantiate.")

    if "params" in config:
        return get_obj_from_str(config["target"])(**config.get("params", dict()))
    else:
        return get_obj_from_str(config["target"])()


def load_config(filename: str):
    """加载参数文件"""

    filename = osp.abspath(filename)
    config = OmegaConf.load(filename)

    Logger.info(f"load_config: {filename}")
    # merge config
    if hasattr(config, "base"):
        base_filename = osp.join(osp.dirname(filename), config.base)
        delattr(config, "base")

        base_config = load_config(base_filename)
        config = OmegaConf.merge(base_config, config)
    return config


def save_config(filename: str, config):
    """保存配置文件"""
    OmegaConf.save(config, filename)
    Logger.info(f"save_config: {filename}")


def make_config():
    """生成配置文件模板"""
    src = osp.join(osp.dirname(__file__), "template.yaml")
    dst = "template.yaml"
    shutil.copy(src, dst)
    Logger.info(f"create: {dst}")


def get_log_dir(config_filename: str):
    """获取log目录"""
    return config_filename.replace("configs", "logs").replace(".yaml", "")


def get_checkpoint_dir(config_filename):
    return osp.join(get_log_dir(config_filename), "checkpoint")


def omegaconf_to_dict(cfg):
    d = dict()
    for key, value in cfg.items():
        d[key] = value
    return d



import argparse
import cProfile
import os
import os.path as osp
import shutil
import sys

import pytorch_lightning as pl
from pytorch_lightning import seed_everything
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.plugins import DDPPlugin
from pytorch_lightning.trainer import Trainer

from ..io.io import mkdir
from ..tools.system_info import get_cpu_info, get_gpu_info, get_package_info
from ..tools import Logger, get_time_str
from .config import (get_checkpoint_dir, get_log_dir, instantiate_from_config,
                     load_config, save_config)
from .dataset import LightningDataset
from .model import LightningModel
from .train_tips import get_train_tips


def train(config_filename: str) -> None:
    """从配置文件启动训练

    Args:
        config_filename (str): 配置文件
    """

    root_path = osp.abspath(config_filename).split("configs")[0]
    sys.path.insert(0, root_path)
    Logger.info(f"sys.path.insert(0, {root_path})")

    config = load_config(config_filename)

    # log_dir
    log_dir = get_log_dir(config_filename)
    if osp.exists(log_dir):
        Logger.error("please create a new config!")
        return
    mkdir(log_dir)
    environments_filename = osp.join(root_path, log_dir, "environments.yaml")
    os.system(f"conda env export > {environments_filename}")
    Logger.logfile(osp.join(log_dir, "train.log"), clear=True)
    Logger.info(f"mkdir: {log_dir}")
    if hasattr(config, "target"):
        Logger.info(f"实验目的[target]: {config.target}")

    # copy python file
    copy_file_list = config.copy_file_list
    if copy_file_list is not None:
        for filename in copy_file_list:
            full_filename = osp.join(root_path, filename)
            if osp.exists(full_filename):
                dst = osp.join(log_dir, "code", filename)
                mkdir(osp.dirname(dst))
                shutil.copy(src=full_filename, dst=dst)
            else:
                Logger.warn(f"not exists: {full_filename}")

    # config
    new_config = osp.join(log_dir, "config.yaml")
    save_config(new_config, config)

    # log file
    Logger.info(f"machine: {config.machine}")
    Logger.info(f"train: {get_time_str()}")
    Logger.info(f"config: {config_filename}")

    # gpu, cpu, package info
    for key, value in get_gpu_info().items():
        Logger.info(f"{key}: {value}")
    for key, value in get_cpu_info().items():
        Logger.info(f"{key}: {value}")
    for key, value in get_package_info().items():
        Logger.info(f"{key}: {value}")

    # seed
    seed_everything(config.seed)
    Logger.info(f"seed:{config.seed}")

    Logger.info(f"gpus:{config.gpus}")
    Logger.info(f"precision:{config.precision}")
    Logger.info(f"train batch_size:{config.data.train.data_loader.batch_size}")
    Logger.info(f"train num_workers:{config.data.train.data_loader.num_workers}")
    Logger.info(f"save_top_k: {config.checkpoint.save_top_k}")
    Logger.info(f"save_last: {config.checkpoint.save_last}")

    resume_from_checkpoint = config.checkpoint.resume_from
    if resume_from_checkpoint:
        Logger.warn(f"resume from:{resume_from_checkpoint}")
    else:
        resume_from_checkpoint = None

    path, version = osp.split(log_dir)
    path, task_name = osp.split(path)
    path, save_dir = osp.split(path)
    if save_dir != "logs":
        save_dir = ""

    trainer = Trainer(
        precision=config.precision,
        gpus=config.gpus,
        sync_batchnorm=True,
        check_val_every_n_epoch=config.check_val_every_n_epoch,
        accumulate_grad_batches=config.accumulate_grad_batches,
        max_epochs=config.max_epochs,
        logger=[
            pl.loggers.TensorBoardLogger(
                save_dir=save_dir,
                name=task_name,
                version=version,
            ),
        ],
        callbacks=[
            LearningRateMonitor(logging_interval="step"),
            ModelCheckpoint(
                dirpath=get_checkpoint_dir(config_filename),
                monitor="eval_loss",
                filename="{epoch:03d}_{eval_loss:.5f}",
                save_last=config.checkpoint.save_last,
                save_top_k=config.checkpoint.save_top_k,
            ),
        ],
        plugins=DDPPlugin(find_unused_parameters=config.find_unused_parameters),
        resume_from_checkpoint=resume_from_checkpoint,
        log_every_n_steps=config.log_every_n_steps,
        profiler="simple",
    )

    Logger.info("prepare to build model...")
    model = instantiate_from_config(config.model)
    lightning_model = LightningModel(
        model=model, optimizer_cfg=config.optimizer, scheduler_cfg=config.scheduler
    )

    Logger.info("prepare to build dataset...")
    data = LightningDataset(**config.data)
    
    Logger.info("prepare to train...")
    Logger.warn(f"Please Run: tensorboard --logdir={osp.join(save_dir, task_name)}")

    if config.cProfile_enable:
        Logger.info(f"cProfile_enable: {config.cProfile_enable}")
        if config.max_epochs != 1:
            Logger.warn(f"Proposed max_epochs=1, now max_epochs={config.max_epochs}")
        profiler = cProfile.Profile()
        profiler.enable()

    trainer.fit(lightning_model, data)

    if config.cProfile_enable:
        profiler.disable()
        filename = osp.join(log_dir, "out.prof")
        profiler.dump_stats(filename)
        Logger.info(f"save: {filename}")
        Logger.warn(f"Please Run: snakeviz {filename}")
        svg = filename.replace(".prof", "svg")
        Logger.warn(f"Please Run: flameprof {filename} > {svg}")


def train_main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "config", default="config.yaml", type=str, help="config filename"
    )

    args = parser.parse_args()

    if args.config == "tips":
        for cnt, tip in enumerate(get_train_tips()):
            title = tip["title"]
            url = tip["url"]
            Logger.warn(f"Tip {cnt+1}:")
            Logger.info(f"title: {title}")
            Logger.info(f"url: {url}")
            print()
        return
    elif not osp.exists(args.config):
        Logger.error(f"not exists: {args.config}")
        sys.exit(0)

    train(args.config)



"""base dataset"""
import os.path as osp
from typing import List

import cv2
from torch.utils.data import Dataset


class BaseDataset(Dataset):
    """数据集模板类"""

    def info(self) -> dict:
        """数据集相关信息"""
        raise NotImplementedError

    def keys(self) -> List[str]:
        """数据集包含的内容"""
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __getitem__(self, index: int) -> dict:
        raise NotImplementedError



import numpy as np


def get_center_scale(box_info):
    x, y, w, h = box_info

    center = np.zeros((2), dtype=np.float32)
    center[0] = x + w * 0.5
    center[1] = y + h * 0.5

    scale = np.array([w * 1.0, h * 1.0], dtype=np.float32)

    return center, scale


def get_bbox(joint_img):
    x_img, y_img = joint_img[:, 0], joint_img[:, 1]
    xmin = min(x_img)
    ymin = min(y_img)
    xmax = max(x_img)
    ymax = max(y_img)

    x_center = (xmin + xmax) / 2.0
    width = xmax - xmin
    xmin = x_center - 0.5 * width  # * 1.2
    xmax = x_center + 0.5 * width  # * 1.2

    y_center = (ymin + ymax) / 2.0
    height = ymax - ymin
    ymin = y_center - 0.5 * height  # * 1.2
    ymax = y_center + 0.5 * height  # * 1.2

    bbox = np.array([xmin, ymin, xmax - xmin, ymax - ymin]).astype(np.float32)
    return bbox


def process_bbox(bbox, aspect_ratio=None, scale=1.0):
    # sanitize bboxes
    x, y, w, h = bbox
    x1, y1, x2, y2 = x, y, x + (w - 1), y + (h - 1)
    if w * h > 0 and x2 >= x1 and y2 >= y1:
        bbox = np.array([x1, y1, x2 - x1, y2 - y1])
    else:
        return None
    return bbox
    # aspect ratio preserving bbox
    w = bbox[2]
    h = bbox[3]
    c_x = bbox[0] + w / 2.0
    c_y = bbox[1] + h / 2.0
    if aspect_ratio is None:
        aspect_ratio = cfg.MODEL.input_shape[1] / cfg.MODEL.input_shape[0]
    if w > aspect_ratio * h:
        h = w / aspect_ratio
    elif w < aspect_ratio * h:
        w = h * aspect_ratio
    bbox[2] = w * scale  # *1.25
    bbox[3] = h * scale  # *1.25
    bbox[0] = c_x - bbox[2] / 2.0
    bbox[1] = c_y - bbox[3] / 2.0
    return bbox


def cam2pixel(cam_coord, f, c):
    x = cam_coord[:, 0] / cam_coord[:, 2] * f[0] + c[0]
    y = cam_coord[:, 1] / cam_coord[:, 2] * f[1] + c[1]
    z = cam_coord[:, 2] / cam_coord[:, 2]
    img_coord = np.concatenate((x[:, None], y[:, None], z[:, None]), axis=1)
    return img_coord


def world2cam(world_coord, R, t):
    cam_coord = np.dot(R, world_coord.transpose(1, 0)).transpose(1, 0) + t.reshape(1, 3)
    return cam_coord


def pixel2cam(coords, c, f):
    cam_coord = np.zeros((len(coords), 3))
    z = coords[..., 2].reshape(-1, 1)

    cam_coord[..., :2] = (coords[..., :2] - c) * z / f
    cam_coord[..., 2] = coords[..., 2]

    return cam_coord



def rigid_transform_3D(A, B):
    n, dim = A.shape
    centroid_A = np.mean(A, axis = 0)
    centroid_B = np.mean(B, axis = 0)
    H = np.dot(np.transpose(A - centroid_A), B - centroid_B) / n
    U, s, V = np.linalg.svd(H)
    R = np.dot(np.transpose(V), np.transpose(U))
    if np.linalg.det(R) < 0:
        s[-1] = -s[-1]
        V[2] = -V[2]
        R = np.dot(np.transpose(V), np.transpose(U))

    varP = np.var(A, axis=0).sum()
    c = 1/varP * np.sum(s)

    t = -np.dot(c*R, np.transpose(centroid_A)) + np.transpose(centroid_B)
    return c, R, t


def rigid_align(A, B):
    c, R, t = rigid_transform_3D(A, B)
    A2 = np.transpose(np.dot(c*R, np.transpose(A))) + t
    return A2


"""md5sum

"""
import hashlib
import os.path as osp
from typing import List, Tuple

from ..io.io import find_all_file
from .tasks import multi_tasks


def md5sum(filename: str) -> str:
    """计算文件的md5值

    copy from https://www.aiuai.cn/aifarm1114.html
    """
    f = open(filename, "rb")
    md5 = hashlib.md5()
    while True:
        fb = f.read(8096)
        if not fb:
            break
        md5.update(fb)
    f.close()
    return md5.hexdigest()


def md5sum_folder(folder: str) -> List[Tuple[str, str]]:
    """计算文件夹的md5值(支持多级目录)

    Args:
        folder (str): 文件夹

    Returns:
        List[Tuple[str, str]]: [[md5, filename],...]
    """
    md5_list = []
    filename_list = list(find_all_file(folder))
    md5_list = multi_tasks(md5sum, filename_list, pbar_name="md5sum", workers=12)
    return zip(md5_list, filename_list)


def save_md5sum(path: str, md5_filename: str = "md5.txt", cut_folder: str = "") -> None:
    """保存MD5到文件

    Args:
        path (str): file or folder
        md5_filename (str): md5.txt
        cut_folder (str): 去除路径头部
    """
    if osp.isfile(path):
        md5_filename_list = [[md5sum(path), path]]
    elif osp.isdir(path):
        md5_filename_list = md5sum_folder(path)
    else:
        assert False, "path is bad"

    if not cut_folder.endswith("/"):
        cut_folder = cut_folder + "/"
    with open(md5_filename, "w", encoding="utf-8") as fw:
        for md5, filename in md5_filename_list:
            if cut_folder and filename.startswith(cut_folder):
                filename = filename[len(cut_folder) :]
            fw.write(f"{md5} {filename}\n")


def check_md5sum(path: str, md5_filename: str = "md5.txt") -> None:
    """检查MD5

    Args:
        path (str): folder
        md5_filename (str): md5.txt
    """
    md5_list = []
    with open(md5_filename, "r") as fr:
        for line in fr.readlines():
            md5, filename = line.replace("  ", " ").strip().split(" ")
            md5_list.append([md5, filename])

    def check(md5, filename):
        if md5sum(osp.join(path, filename)) != md5:
            print("MD5 bad:", filename)

    multi_tasks(check, md5_list, pbar_name="check md5")



import os
import os.path as osp
from collections import OrderedDict

import torch
import torch.nn as nn

from .config import get_log_dir, instantiate_from_config, load_config

__all__ = ["load_checkpoint", "get_best_checkpoint", "get_model"]


def load_checkpoint(model: nn.Module, checkpoint_file: str, remove_head: bool = True) -> nn.Module:
    """load checkpoints

    Args:
        model (nn.Module): model
        checkpoint_file (str): checkpoint filename
        remove_head (bool, optional): 是否去除[model.]. Defaults to True.

    Returns:
        nn.Module: model
    """
    state_dict = torch.load(checkpoint_file)["state_dict"]
    new_state_dict = dict()
    for name, value in state_dict.items():
        if remove_head and name[:6] == "model.":
            new_name = name[6:]
        else:
            new_name = name
        new_state_dict[new_name] = value
    model.load_state_dict(new_state_dict)
    return model


def get_best_checkpoint_from_folder(folder: str, select: str = "min") -> str:
    """从指定文件夹中获取最好的模型文件

    Args:
        folder (str): 指定文件夹
        select (str, optional): min or max. Defaults to "min".

    Returns:
        str: 最好的模型文件
    """
    assert select in ["min", "max"]

    name_list = os.listdir(folder)
    if "last.ckpt" in name_list:
        name_list.remove("last.ckpt")

    best_score = None
    best_name = None
    for name in name_list:
        score = float(name.split("=")[-1].replace(".ckpt", ""))
        if best_score is None:
            best_score, best_name = score, name

        if select == "min" and score < best_score:
            best_score, best_name = score, name
        elif select == "max" and score > best_score:
            best_score, best_name = score, name

    if best_name is None:
        return None

    return osp.join(folder, best_name)


def get_best_checkpoint(config_filename: str, select: str = "min") -> str:
    """获取当前配置下最好的模型文件

    Args:
        config_filename (str): 配置文件
        select (str, optional): min or max. Defaults to "min".

    Returns:
        str: 最好的模型文件
    """
    checkpoint_folder = osp.join(get_log_dir(config_filename), "checkpoint")
    return get_best_checkpoint_from_folder(checkpoint_folder, select=select)


def get_model(config_filename: str) -> nn.Module:
    """获取当前配置下的模型

    Args:
        config_filename (str): 配置文件

    Returns:
        nn.Module: model
    """
    config = load_config(config_filename)
    model = instantiate_from_config(config.model)
    return model


def convert_ckpt_to_pth(ckpt_filename: str, remove_head: bool = True):
    """将ckpt转换成pytorch可正常load的pth

    Args:
        ckpt_filename (str): ckpt filename
        remove_head (bool, optional): 是否去除[model.]. Defaults to True.
    """
    new_d = dict()
    new_d["state_dict"] = OrderedDict()

    d = torch.load(ckpt_filename)
    for name in d["state_dict"]:
        if remove_head and name[:6] == "model.":
            new_name = name[6:]
        else:
            new_name = name
        new_d["state_dict"][new_name] = d["state_dict"][name]
    torch.save(new_d, ckpt_filename.replace(".ckpt", ".pth"))



import copy
import json
import os
import os.path as osp

import cv2
import numpy as np
import torch
from human_cver.vis.plot_kps import plot_kps_2d
from pycocotools.coco import COCO

from ...geometric.coord import (cam2pixel, get_bbox, process_bbox)
from ...models.smpl.smpl import SMPL


class PW3D(torch.utils.data.Dataset):
    def __init__(self, dataset_path: str, pw3d_path, mode: str = "test", smpl: SMPL = None):
        self.mode = mode  #'validation'
        self.data_path = dataset_path
        self.img_path = osp.join(pw3d_path, "imageFiles")
        self.smpl = smpl
        # SMPL joint set
        # self.mesh_model = SMPL()
        # self.smpl_root_joint_idx = self.mesh_model.root_joint_idx
        # self.face_kps_vertex = self.mesh_model.face_kps_vertex
        self.smpl_vertex_num = 6890
        self.smpl_joint_num = 24
        self.smpl_flip_pairs = ((1, 2), (4, 5), (7, 8), (10, 11), (13, 14), (16, 17), (18, 19), (20, 21), (22, 23))
        self.smpl_skeleton = (
            (0, 1),
            (1, 4),
            (4, 7),
            (7, 10),
            (0, 2),
            (2, 5),
            (5, 8),
            (8, 11),
            (0, 3),
            (3, 6),
            (6, 9),
            (9, 14),
            (14, 17),
            (17, 19),
            (19, 21),
            (21, 23),
            (9, 13),
            (13, 16),
            (16, 18),
            (18, 20),
            (20, 22),
            (9, 12),
            (12, 15),
        )
        # self.joint_regressor_smpl = self.mesh_model.layer['neutral'].th_J_regressor

        # H36M joint set
        self.human36_root_joint_idx = 0
        self.human36_eval_joint = (1, 2, 3, 4, 5, 6, 8, 10, 11, 12, 13, 14, 15, 16)
        self.human36_skeleton = (
            (0, 7),
            (7, 8),
            (8, 9),
            (9, 10),
            (8, 11),
            (11, 12),
            (12, 13),
            (8, 14),
            (14, 15),
            (15, 16),
            (0, 1),
            (1, 2),
            (2, 3),
            (0, 4),
            (4, 5),
            (5, 6),
        )
        # self.joint_regressor_human36 = torch.Tensor(self.mesh_model.joint_regressor_h36m)

        # COCO joint set
        self.coco_joint_num = 19  # 17 + 2, manually added pelvis and neck
        self.coco_joints_name = (
            "Nose",
            "L_Eye",
            "R_Eye",
            "L_Ear",
            "R_Ear",
            "L_Shoulder",
            "R_Shoulder",
            "L_Elbow",
            "R_Elbow",
            "L_Wrist",
            "R_Wrist",
            "L_Hip",
            "R_Hip",
            "L_Knee",
            "R_Knee",
            "L_Ankle",
            "R_Ankle",
            "Pelvis",
            "Neck",
        )
        self.coco_flip_pairs = ((1, 2), (3, 4), (5, 6), (7, 8), (9, 10), (11, 12), (13, 14), (15, 16))
        self.coco_skeleton = (
            (1, 2),
            (0, 1),
            (0, 2),
            (2, 4),
            (1, 3),
            (6, 8),
            (8, 10),
            (5, 7),
            (7, 9),
            (12, 14),
            (14, 16),
            (11, 13),
            (13, 15),  # (5, 6), (11, 12))
            (17, 11),
            (17, 12),
            (17, 18),
            (18, 5),
            (18, 6),
            (18, 0),
        )
        # self.joint_regressor_coco = torch.Tensor(self.mesh_model.joint_regressor_coco)
        self.openpose_joints_name = (
            "Nose",
            "Neck",
            "R_Shoulder",
            "R_Elbow",
            "R_Wrist",
            "L_Shoulder",
            "L_Elbow",
            "L_Wrist",
            "R_Hip",
            "R_Knee",
            "R_Ankle",
            "L_Hip",
            "L_Knee",
            "L_Ankle",
            "R_Eye",
            "L_Eye",
            "R_Ear",
            "L_Ear",
            "Pelvis",
        )

        input_joint_name = "coco"
        self.joint_num, self.skeleton, self.flip_pairs = self.get_joint_setting(input_joint_name)

        self.datalist, self.video_indices = self.load_data()  # self.video_indexes: 37 video, and indices of each video

        print("3dpw data len: ", len(self.datalist))

    def get_joint_setting(self, joint_category="human36"):
        joint_num = eval(f"self.{joint_category}_joint_num")
        skeleton = eval(f"self.{joint_category}_skeleton")
        flip_pairs = eval(f"self.{joint_category}_flip_pairs")

        return joint_num, skeleton, flip_pairs

    def get_smpl_coord(self, smpl_param):
        pose, shape, trans, gender = smpl_param["pose"], smpl_param["shape"], smpl_param["trans"], smpl_param["gender"]
        smpl_pose = torch.FloatTensor(pose).view(-1, 3)
        smpl_shape = torch.FloatTensor(shape).view(1, -1)
        # translation vector from smpl coordinate to 3dpw world coordinate
        smpl_trans = torch.FloatTensor(trans).view(-1, 3)

        smpl_pose = smpl_pose.view(1, -1)
        # get mesh and joint coordinates
        smpl_mesh_coord, smpl_joint_coord = self.smpl(pose=smpl_pose, betas=smpl_shape, gender=gender)  # , smpl_trans)

        # incorporate face keypoints
        smpl_mesh_coord = smpl_mesh_coord.numpy().astype(np.float32).reshape(-1, 3)
        smpl_joint_coord = smpl_joint_coord.numpy().astype(np.float32).reshape(-1, 3)

        smpl_mesh_coord += smpl_trans.numpy()
        smpl_joint_coord += smpl_trans.numpy()
        # meter -> milimeter
        smpl_mesh_coord *= 1000
        smpl_joint_coord *= 1000
        return smpl_mesh_coord, smpl_joint_coord

    def load_data(self):
        print("Load annotations of 3DPW ")
        db = COCO(osp.join(self.data_path, "3DPW_latest_" + self.mode + ".json"))

        # get detected 2d pose
        with open(osp.join(self.data_path, f"darkpose_3dpw_{self.mode}set_output.json")) as f:  #
            pose2d_outputs = {}
            data = json.load(f)
            for item in data:
                annot_id = str(item["annotation_id"])
                pose2d_outputs[annot_id] = {"coco_joints": np.array(item["keypoints"], dtype=np.float32)[:, :3]}

        datalist = []
        custompose_count = 0
        for aid in db.anns.keys():
            aid = int(aid)
            ann = db.anns[aid]
            image_id = ann["image_id"]

            img = db.loadImgs(image_id)[0]
            img_width, img_height = img["width"], img["height"]
            sequence_name = img["sequence"]
            img_name = img["file_name"]

            img_path = osp.join(self.img_path, sequence_name, img_name)
            cam_param = {k: np.array(v, dtype=np.float32) for k, v in img["cam_param"].items()}

            smpl_param = ann["smpl_param"]
            pid = ann["person_id"]
            vid_name = sequence_name + str(pid)
            bbox = process_bbox(np.array(ann["bbox"]))
            if bbox is None:
                continue
            
            # 18, 3
            openpose = np.array(ann["openpose_result"], dtype=np.float32).reshape(-1, 3)
            openpose = self.add_pelvis_and_neck(openpose, self.openpose_joints_name, only_pelvis=True)

            # 17, 3
            custompose = np.array(pose2d_outputs[str(aid)]["coco_joints"])
            custompose = self.add_pelvis_and_neck(custompose, self.coco_joints_name)
            custompose_count += 1

            datalist.append(
                {
                    "annot_id": aid,
                    "person_id": pid,
                    "image_id": image_id,
                    "img_path": img_path,
                    "vid_name": vid_name,
                    "img_shape": (img_height, img_width),
                    "cam_param": cam_param,
                    "bbox": bbox,
                    "smpl_param": smpl_param,
                    "pred_pose2d": custompose,
                }
            )

        datalist = sorted(datalist, key=lambda x: (x["person_id"], x["img_path"]))
        valid_names = np.array([data["vid_name"] for data in datalist])
        unique_names = np.unique(valid_names)
        video_indices = []
        for u_n in unique_names:
            indexes = valid_names == u_n
            video_indices.append(indexes)

        print("num custom pose: ", custompose_count)
        return datalist, video_indices

    def add_pelvis_and_neck(self, joint_coord, joints_name, only_pelvis=False):
        lhip_idx = joints_name.index("L_Hip")
        rhip_idx = joints_name.index("R_Hip")
        pelvis = (joint_coord[lhip_idx, :] + joint_coord[rhip_idx, :]) * 0.5
        pelvis = pelvis.reshape((1, -1))

        lshoulder_idx = joints_name.index("L_Shoulder")
        rshoulder_idx = joints_name.index("R_Shoulder")
        neck = (joint_coord[lshoulder_idx, :] + joint_coord[rshoulder_idx, :]) * 0.5
        neck = neck.reshape((1, -1))

        if only_pelvis:
            joint_coord = np.concatenate((joint_coord, pelvis))
        else:
            joint_coord = np.concatenate((joint_coord, pelvis, neck))
        return joint_coord

    def __len__(self):
        return len(self.datalist)

    def render(self, index):
        d = self.__getitem__(index)

        img_path = d["img_path"]
        img = cv2.imread(img_path)

        cam_param = d["cam_param"]
        smpl_param = d["smpl_param"]
        mesh_cam, joint_cam_smpl = self.get_smpl_coord(smpl_param)
        mesh_coord = cam2pixel(mesh_cam, f=cam_param["focal"], c=cam_param["princpt"])
        plot_kps_2d(img, mesh_coord)

        return {
            "img": img,
            "pose_3d": joint_cam_smpl
        }

    def __getitem__(self, idx):
        return copy.deepcopy(self.datalist[idx])




import os.path as osp
import numpy as np
import math
import torch
import json
import copy
import transforms3d
import scipy.sparse
import cv2
from pycocotools.coco import COCO

# from core.config import cfg 
# from graph_utils import build_coarse_graphs
# from noise_utils import synthesize_pose

# from smpl import SMPL
from ....geometric.coord import world2cam, cam2pixel, process_bbox, rigid_align, get_bbox
# from aug_utils import affine_transform, j2d_processing, augm_params, j3d_processing, flip_2d_joint
from .noise_stats import error_distribution
from ....models.smpl.smpl import SMPL
from ....models.smpl.kp_map import SMPL_ROOT_IDX
# from funcs_utils import save_obj, stop
# from vis import vis_3d_pose, vis_2d_pose
from ....vis.plot_kps import plot_kps_2d

class Human36M(torch.utils.data.Dataset):
    r""" Human3.6m dataset

    Args:
        dataset_root: str
        mode: train or test
        smpl: used to render smpl mesh on image
    """
    def __init__(self, dataset_root:str="/media/lwk/T7/datasets/Pose2Mesh/Human3.6M", mode:str="train", smpl:SMPL=None):
        self.dataset_root = dataset_root
        self.debug = True
        self.data_split = mode
        self.img_dir = osp.join(dataset_root, 'images')
        self.annot_path = osp.join(dataset_root, 'annotations')
        self.subject_genders = {1: 'female', 5: 'female', 6: 'male', 7: 'female', 8: 'male', 9: 'male', 11: 'male'}
        self.protocol = 2
        self.action_name = ['Directions', 'Discussion', 'Eating', 'Greeting', 'Phoning', 'Posing', 'Purchases',
                            'Sitting', 'SittingDown', 'Smoking', 'Photo', 'Waiting', 'Walking', 'WalkDog',
                            'WalkTogether']
        self.fitting_thr = 25  # milimeter
        self.smpl = smpl
        if smpl is not None:
            assert smpl.kps_num() == 24

        # SMPL joint set
        # self.mesh_model = SMPL()
        # self.smpl_root_joint_idx = SMPL_ROOT_IDX
        self.smpl_root_joint_idx = 0
        # self.smpl_face_kps_vertex = self.mesh_model.face_kps_vertex
        self.smpl_vertex_num = 6890
        self.smpl_joint_num = 24
        self.smpl_flip_pairs = ((1, 2), (4, 5), (7, 8), (10, 11), (13, 14), (16, 17), (18, 19), (20, 21), (22, 23))
        self.smpl_skeleton = (
            (0, 1), (1, 4), (4, 7), (7, 10), (0, 2), (2, 5), (5, 8), (8, 11), (0, 3), (3, 6), (6, 9), (9, 14), (14, 17),
            (17, 19), (19, 21), (21, 23), (9, 13), (13, 16), (16, 18), (18, 20), (20, 22), (9, 12), (12, 15))
        # self.joint_regressor_smpl = self.mesh_model.layer['neutral'].th_J_regressor

        # H36M joint set
        self.human36_joint_num = 17
        self.human36_joints_name = (
        'Pelvis', 'R_Hip', 'R_Knee', 'R_Ankle', 'L_Hip', 'L_Knee', 'L_Ankle', 'Torso', 'Neck', 'Nose', 'Head',
        'L_Shoulder', 'L_Elbow', 'L_Wrist', 'R_Shoulder', 'R_Elbow', 'R_Wrist')
        self.human36_flip_pairs = ((1, 4), (2, 5), (3, 6), (14, 11), (15, 12), (16, 13))
        self.human36_skeleton = (
        (0, 7), (7, 8), (8, 9), (9, 10), (8, 11), (11, 12), (12, 13), (8, 14), (14, 15), (15, 16), (0, 1), (1, 2),
        (2, 3), (0, 4), (4, 5), (5, 6))
        self.human36_root_joint_idx = self.human36_joints_name.index('Pelvis')
        self.human36_error_distribution = self.get_stat()
        self.human36_eval_joint = (1, 2, 3, 4, 5, 6, 8, 10, 11, 12, 13, 14, 15, 16)
        # self.joint_regressor_human36 = self.mesh_model.joint_regressor_h36m

        # COCO joint set
        self.coco_joint_num = 19  # 17 + 2, manually added pelvis and neck
        self.coco_joints_name = (
            'Nose', 'L_Eye', 'R_Eye', 'L_Ear', 'R_Ear', 'L_Shoulder', 'R_Shoulder', 'L_Elbow', 'R_Elbow', 'L_Wrist',
            'R_Wrist', 'L_Hip', 'R_Hip', 'L_Knee', 'R_Knee', 'L_Ankle', 'R_Ankle', 'Pelvis', 'Neck')
        self.coco_flip_pairs = ((1, 2), (3, 4), (5, 6), (7, 8), (9, 10), (11, 12), (13, 14), (15, 16))
        self.coco_skeleton = (
            (1, 2), (0, 1), (0, 2), (2, 4), (1, 3), (6, 8), (8, 10), (5, 7), (7, 9), (12, 14), (14, 16), (11, 13),
            (13, 15), #(5, 6), #(11, 12),
            (17, 11), (17, 12), (17, 18), (18, 5), (18, 6), (18, 0))
        # self.joint_regressor_coco = self.mesh_model.joint_regressor_coco

        # self.input_joint_name = cfg.DATASET.input_joint_set  # 'coco'
        # self.joint_num, self.skeleton, self.flip_pairs = self.get_joint_setting(self.input_joint_name)

        self.datalist, skip_idx, skip_img_path = self.load_data()
        if self.data_split == 'test':
            det_2d_data_path = osp.join(dataset_root, 'absnet_output_on_testset.json')
            self.datalist_pose2d_det = self.load_pose2d_det(det_2d_data_path, skip_img_path)
            print("Check lengths of annotation and detection output: ", len(self.datalist), len(self.datalist_pose2d_det))

        # self.graph_Adj, self.graph_L, self.graph_perm, self.graph_perm_reverse = \
        #     build_coarse_graphs(self.mesh_model.face, self.joint_num, self.skeleton, self.flip_pairs, levels=9)

    def load_pose2d_det(self, data_path, skip_list):
        pose_list = []
        with open(data_path) as f:
            data = json.load(f)
            for img_path, pose2d in data.items():
                pose2d = np.array(pose2d, dtype=np.float32)
                if img_path in skip_list:
                    continue
                pose_list.append({'img_name': img_path, 'pose2d': pose2d})
        pose_list = sorted(pose_list, key=lambda x: x['img_name'])
        return pose_list

    def get_joint_setting(self, joint_category='human36'):
        joint_num = eval(f'self.{joint_category}_joint_num')
        skeleton = eval(f'self.{joint_category}_skeleton')
        flip_pairs = eval(f'self.{joint_category}_flip_pairs')

        return joint_num, skeleton, flip_pairs

    def get_subsampling_ratio(self):
        if self.data_split == 'train':
            return 5  # 50
        elif self.data_split == 'test':
            return 50 #
        else:
            assert 0, print('Unknown subset')

    def get_subject(self):
        if self.data_split == 'train':
            if self.protocol == 1:
                subject = [1, 5, 6, 7, 8, 9]
            elif self.protocol == 2:
                subject = [1, 5, 6, 7, 8]
        elif self.data_split == 'test':
            if self.protocol == 1:
                subject = [11]
            elif self.protocol == 2:
                subject = [9, 11]
        else:
            assert 0, print("Unknown subset")

        if self.debug:
            subject = subject[0:1]

        return subject

    def get_stat(self):
        ordered_stats = []
        for joint in self.human36_joints_name:
            item = list(filter(lambda stat: stat['Joint'] == joint, error_distribution))[0]
            ordered_stats.append(item)

        return ordered_stats

    def generate_syn_error(self):
        noise = np.zeros((self.human36_joint_num, 2), dtype=np.float32)
        weight = np.zeros(self.human36_joint_num, dtype=np.float32)
        for i, ed in enumerate(self.human36_error_distribution):
            noise[i, 0] = np.random.normal(loc=ed['mean'][0], scale=ed['std'][0])
            noise[i, 1] = np.random.normal(loc=ed['mean'][1], scale=ed['std'][1])
            weight[i] = ed['weight']

        prob = np.random.uniform(low=0.0, high=1.0, size=self.human36_joint_num)
        weight = (weight > prob)
        noise = noise * weight[:, None]

        return noise

    def load_json(self, subject, key):
        with open(osp.join(self.annot_path, f'Human36M_subject{subject}_{key}.json'), 'r') as f:
            annot = json.load(f)
        return annot

    def load_data(self):
        print('Load annotations of Human36M Protocol ' + str(self.protocol))
        subject_list = self.get_subject()
        sampling_ratio = self.get_subsampling_ratio()

        # aggregate annotations from each subject
        db = COCO()
        cameras = {}
        joints = {}
        smpl_params = {}
        for subject in subject_list:
            # data load
            annot = self.load_json(subject, "data")
            # with open(osp.join(self.annot_path, 'Human36M_subject' + str(subject) + '_data.json'), 'r') as f:
            #     annot = json.load(f)
            if len(db.dataset) == 0:
                for k, v in annot.items():
                    db.dataset[k] = v
            else:
                for k, v in annot.items():
                    db.dataset[k] += v
            # camera load
            cameras[str(subject)] = self.load_json(subject, "camera")
            # with open(osp.join(self.annot_path, 'Human36M_subject' + str(subject) + '_camera.json'), 'r') as f:
                # cameras[str(subject)] = json.load(f)
            # joint coordinate load
            joints[str(subject)] = self.load_json(subject, "joint_3d")
            # with open(osp.join(self.annot_path, 'Human36M_subject' + str(subject) + '_joint_3d.json'), 'r') as f:
            #     joints[str(subject)] = json.load(f)
            # smpl parameter load
            smpl_params[str(subject)] = self.load_json(subject, "smpl_param")
            # with open(osp.join(self.annot_path, 'Human36M_subject' + str(subject) + '_smpl_param.json'), 'r') as f:
            #     smpl_params[str(subject)] = json.load(f)
        db.createIndex()

        skip_idx = []
        datalist = []
        skip_img_idx = []
        for aid in db.anns.keys():
            ann = db.anns[aid]
            image_id = ann['image_id']
            img = db.loadImgs(image_id)[0]
            img_path = osp.join(self.img_dir, img['file_name'])
            img_name = img_path.split('/')[-1]

            # check subject and frame_idx
            frame_idx = img['frame_idx']

            if frame_idx % sampling_ratio != 0:
                continue

            # check smpl parameter exist
            subject = img['subject']
            action_idx = img['action_idx']

            subaction_idx = img['subaction_idx']
            frame_idx = img['frame_idx']
            try:
                smpl_param = smpl_params[str(subject)][str(action_idx)][str(subaction_idx)][str(frame_idx)]
            except KeyError:
                skip_idx.append(image_id)
                skip_img_idx.append(img_path.split('/')[-1])
                continue

            smpl_param['gender'] = 'neutral'  # self.subject_genders[subject] # set corresponding gender

            # camera parameter
            cam_idx = img['cam_idx']
            cam_param = cameras[str(subject)][str(cam_idx)]
            R, t, f, c = np.array(cam_param['R'], dtype=np.float32), np.array(cam_param['t'],
                                                                              dtype=np.float32), np.array(
                cam_param['f'], dtype=np.float32), np.array(cam_param['c'], dtype=np.float32)
            cam_param = {'R': R, 't': t, 'focal': f, 'princpt': c}

            # project world coordinate to cam, image coordinate space
            joint_world = np.array(joints[str(subject)][str(action_idx)][str(subaction_idx)][str(frame_idx)],
                                   dtype=np.float32)
            joint_cam = world2cam(joint_world, R, t)
            joint_img = cam2pixel(joint_cam, f, c)
            joint_vis = np.ones((self.human36_joint_num, 1))

            bbox = process_bbox(np.array(ann['bbox']))
            if bbox is None: continue

            datalist.append({
                'img_path': img_path,
                'img_name': img_name,
                'img_id': image_id,
                'bbox': bbox,
                'img_hw': (img['height'], img['width']),
                'joint_img': joint_img,  # [x_img, y_img, z_cam]
                'joint_cam': joint_cam,  # [X, Y, Z] in camera coordinate
                'joint_vis': joint_vis,
                'smpl_param': smpl_param,
                'cam_param': cam_param})

        datalist = sorted(datalist, key=lambda x: x['img_name'])

        return datalist, skip_idx, skip_img_idx

    def get_smpl_coord(self, smpl_param, cam_param):
        pose, shape, trans, gender = smpl_param['pose'], smpl_param['shape'], smpl_param['trans'], smpl_param['gender']
        # smpl parameters (pose: 72 dimension, shape: 10 dimension)
        smpl_pose = torch.FloatTensor(pose).view(-1, 3)
        smpl_shape = torch.FloatTensor(shape).view(1, -1)
        # translation vector from smpl coordinate to h36m world coordinate
        trans = np.array(trans, dtype=np.float32).reshape(3)
        # camera rotation and translation
        R = np.array(cam_param['R'],dtype=np.float32).reshape(3, 3)
        t = np.array(cam_param['t'],dtype=np.float32).reshape(3)

        # change to mean shape if beta is too far from it
        smpl_shape[(smpl_shape.abs() > 3).any(dim=1)] = 0.

        # transform world coordinate to camera coordinate
        root_pose = smpl_pose[0, :].numpy()
        angle = np.linalg.norm(root_pose)
        root_pose = transforms3d.axangles.axangle2mat(root_pose / angle, angle)
        root_pose = np.dot(R, root_pose)
        axis, angle = transforms3d.axangles.mat2axangle(root_pose)
        root_pose = axis * angle
        smpl_pose[0] = torch.from_numpy(root_pose)

        smpl_pose = smpl_pose.view(1, -1)

        # get mesh and joint coordinates
        # smpl_mesh_coord, smpl_joint_coord = self.mesh_model.layer[gender](smpl_pose, smpl_shape)
        smpl_mesh_coord, smpl_joint_coord = self.smpl(pose=smpl_pose, betas=smpl_shape)

        # incorporate face keypoints
        smpl_mesh_coord = smpl_mesh_coord.numpy().astype(np.float32).reshape(-1, 3)
        smpl_joint_coord = smpl_joint_coord.numpy().astype(np.float32).reshape(-1, 3)
        # smpl_face_kps_coord = smpl_mesh_coord[self.face_kps_vertex, :].reshape(-1, 3)
        # smpl_joint_coord = np.concatenate((smpl_joint_coord, smpl_face_kps_coord))

        # compenstate rotation (translation from origin to root joint was not cancled)
        smpl_trans = np.array(trans, dtype=np.float32).reshape(
            3)  # translation vector from smpl coordinate to h36m world coordinate
        smpl_trans = np.dot(R, smpl_trans[:, None]).reshape(1, 3) + t.reshape(1, 3) / 1000
        # root_joint_coord = smpl_joint_coord[self.smpl_root_joint_idx].reshape(1, 3)

        root_joint_coord = smpl_joint_coord[SMPL_ROOT_IDX].reshape(1, 3)
        smpl_trans = smpl_trans - root_joint_coord + np.dot(R, root_joint_coord.transpose(1, 0)).transpose(1, 0)

        # translation
        smpl_mesh_coord += smpl_trans
        smpl_joint_coord += smpl_trans

        # meter -> milimeter
        smpl_mesh_coord *= 1000
        smpl_joint_coord *= 1000

        return smpl_mesh_coord, smpl_joint_coord

    # def get_fitting_error(self, h36m_joint, smpl_mesh):
    #     h36m_joint = h36m_joint - h36m_joint[self.human36_root_joint_idx,None,:] # root-relative

    #     h36m_from_smpl = np.dot(self.joint_regressor_human36, smpl_mesh)
    #     # translation alignment
    #     h36m_from_smpl = h36m_from_smpl - np.mean(h36m_from_smpl,0)[None,:] + np.mean(h36m_joint,0)[None,:]
    #     error = np.sqrt(np.sum((h36m_joint - h36m_from_smpl)**2,1)).mean()
    #     return error

    def get_coco_from_mesh(self, mesh_coord_cam, cam_param):
        # regress coco joints
        joint_coord_cam = np.dot(self.joint_regressor_coco, mesh_coord_cam)
        joint_coord_cam = self.add_pelvis_and_neck(joint_coord_cam)
        # projection
        f, c = cam_param['focal'], cam_param['princpt']
        joint_coord_img = cam2pixel(joint_coord_cam, f, c)

        joint_coord_img[:, 2] = 1
        return joint_coord_cam, joint_coord_img

    def add_pelvis_and_neck(self, joint_coord):
        lhip_idx = self.coco_joints_name.index('L_Hip')
        rhip_idx = self.coco_joints_name.index('R_Hip')
        pelvis = (joint_coord[lhip_idx, :] + joint_coord[rhip_idx, :]) * 0.5
        pelvis = pelvis.reshape((1, -1))

        lshoulder_idx = self.coco_joints_name.index('L_Shoulder')
        rshoulder_idx = self.coco_joints_name.index('R_Shoulder')
        neck = (joint_coord[lshoulder_idx, :] + joint_coord[rshoulder_idx, :]) * 0.5
        neck = neck.reshape((1,-1))

        joint_coord = np.concatenate((joint_coord, pelvis, neck))
        return joint_coord

    def __len__(self):
        return len(self.datalist)

    def render(self, index):
        d = self.__getitem__(index)
        # print(d["img_path"])
        # print(d["img_name"])
        imgname = d["img_path"]
        imgname=imgname.replace("images", "images/images")
        print("imgname:", imgname)
        if not osp.exists(imgname):
            return np.zeros([1000, 1000, 3], dtype=np.uint8)
        img=  cv2.imread(imgname)
        print(d["joint_img"])
        plot_kps_2d(img, d["joint_img"])

        mesh_cam = d["mesh_cam"]
        cam_param = d["cam_param"]
        f = cam_param["focal"]
        c = cam_param["princpt"]
        mesh_cam /= mesh_cam[:, -1:]
        u = f[0] * mesh_cam[:, 0] + c[0]
        v = f[1] * mesh_cam[:, 1] + c[1]
        p2d = np.concatenate([u[:, None], v[:, None]], axis=-1)
        print(cam_param)
        print(p2d)
        plot_kps_2d(img, p2d)
        return img
    def __getitem__(self, idx):
        data = copy.deepcopy(self.datalist[idx])

        img_path = data["img_path"]
        img_name = data["img_name"]
        # print(img_path, img_name, data["img_id"])
        # return data
        img_id, bbox, smpl_param, cam_param, img_shape = data['img_id'], data['bbox'].copy(), data['smpl_param'].copy(), data['cam_param'].copy(), data['img_hw']
        # flip, rot = augm_params(is_train=(self.data_split == 'train'))

        # smpl coordinates
        mesh_cam, joint_cam_smpl = self.get_smpl_coord(smpl_param, cam_param)
        # data["joint_cam_smpl"] = joint_cam_smpl
        data["mesh_cam"] = mesh_cam
        data["joint_cam_smpl"] = joint_cam_smpl
        return data
        # regress coco joints
        joint_cam_coco, joint_img_coco = self.get_coco_from_mesh(mesh_cam, cam_param)
        # h36m joints from datasets
        joint_cam_h36m, joint_img_h36m = data['joint_cam'], data['joint_img'][:, :2]

        # root relative camera coordinate
        mesh_cam = mesh_cam - joint_cam_h36m[:1]
        # joint_cam_smpl = joint_cam_smpl - joint_cam_h36m[:1]
        joint_cam_coco = joint_cam_coco - joint_cam_coco[-2:-1]
        joint_cam_h36m = joint_cam_h36m - joint_cam_h36m[:1]

        # joint_cam is PoseNet target
        if self.input_joint_name == 'coco':
            joint_img, joint_cam = joint_img_coco, joint_cam_coco
        elif self.input_joint_name == 'human36':
            joint_img, joint_cam = joint_img_h36m, joint_cam_h36m

        # make new bbox
        tight_bbox = get_bbox(joint_img)
        bbox = process_bbox(tight_bbox.copy())

        # aug
        joint_img, trans = j2d_processing(joint_img.copy(), (cfg.MODEL.input_shape[1], cfg.MODEL.input_shape[0]), bbox, rot, 0, None)
        if not cfg.DATASET.use_gt_input:
            joint_img = self.replace_joint_img(idx, img_id, joint_img, tight_bbox, trans)
        if flip:
            joint_img = flip_2d_joint(joint_img, cfg.MODEL.input_shape[1], self.flip_pairs)
        joint_cam = j3d_processing(joint_cam, rot, flip, self.flip_pairs)


        # vis
        # img = cv2.imread(img_path, cv2.IMREAD_COLOR | cv2.IMREAD_IGNORE_ORIENTATION)
        # new_img = cv2.warpAffine(img, trans, (cfg.MODEL.input_shape[1], cfg.MODEL.input_shape[0]), flags=cv2.INTER_LINEAR)
        # vis_2d_pose(joint_img, new_img, self.human36_skeleton, prefix='detection')
        # vis_3d_pose(joint_cam, self.human36_skeleton, joint_set_name='human36', gt=True)

        #  -> 0~1
        joint_img = joint_img[:, :2]
        joint_img /= np.array([[cfg.MODEL.input_shape[1], cfg.MODEL.input_shape[0]]])

        # normalize loc&scale
        mean, std = np.mean(joint_img, axis=0), np.std(joint_img, axis=0)
        joint_img = (joint_img.copy() - mean) / std

        if cfg.MODEL.name == 'pose2mesh_net':
            # default valid
            mesh_valid = np.ones((len(mesh_cam), 1), dtype=np.float32)
            reg_joint_valid = np.ones((len(joint_cam_h36m), 1), dtype=np.float32)
            lift_joint_valid = np.ones((len(joint_cam), 1), dtype=np.float32)
            # if fitted mesh is too far from h36m gt, discard it
            error = self.get_fitting_error(joint_cam_h36m, mesh_cam)
            if error > self.fitting_thr:
                mesh_valid[:] = 0
                if self.input_joint_name == 'coco':
                    lift_joint_valid[:] = 0

            inputs = {'pose2d': joint_img}
            targets = {'mesh': mesh_cam / 1000, 'lift_pose3d': joint_cam, 'reg_pose3d': joint_cam_h36m}
            meta = {'mesh_valid': mesh_valid, 'lift_pose3d_valid': lift_joint_valid, 'reg_pose3d_valid': reg_joint_valid}

            return inputs, targets, meta

        elif cfg.MODEL.name == 'posenet':
            # default valid
            joint_valid = np.ones((len(joint_cam), 1), dtype=np.float32)

            # if fitted mesh is too far from h36m gt, discard it
            if self.input_joint_name == 'coco':
                error = self.get_fitting_error(joint_cam_h36m, mesh_cam)
                if (error > self.fitting_thr):
                    joint_valid[:] = 0

            return joint_img, joint_cam, joint_valid

    # def replace_joint_img(self, idx, img_id, joint_img, bbox, trans):
    #     if self.input_joint_name == 'coco':
    #         joint_img_coco = joint_img
    #         if self.data_split == 'train':
    #             xmin, ymin, xmax, ymax = bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]
    #             pt1 = affine_transform(np.array([xmin, ymin]), trans)
    #             pt2 = affine_transform(np.array([xmax, ymin]), trans)
    #             pt3 = affine_transform(np.array([xmax, ymax]), trans)
    #             area = math.sqrt(pow(pt2[0] - pt1[0], 2) + pow(pt2[1] - pt1[1], 2)) * math.sqrt(
    #                 pow(pt3[0] - pt2[0], 2) + pow(pt3[1] - pt2[1], 2))
    #             joint_img_coco[:17, :] = synthesize_pose(joint_img_coco[:17, :], area, num_overlap=0)
    #             return joint_img_coco
    #         else:
    #             joint_img_coco = self.datalist_pose2d_det[img_id]
    #             joint_img_coco = self.add_pelvis_and_neck(joint_img_coco)
    #             for i in range(self.coco_joint_num):
    #                 joint_img_coco[i, :2] = affine_transform(joint_img_coco[i, :2].copy(), trans)
    #             return joint_img_coco

    #     elif self.input_joint_name == 'human36':
    #         joint_img_h36m = joint_img
    #         if self.data_split == 'train':
    #             joint_syn_error = (self.generate_syn_error() / 256) * np.array(
    #                 [cfg.MODEL.input_shape[1], cfg.MODEL.input_shape[0]], dtype=np.float32)
    #             joint_img_h36m = joint_img_h36m[:, :2] + joint_syn_error
    #             return joint_img_h36m
    #         else:
    #             det_data = self.datalist_pose2d_det[idx]
    #             # assert img_name == det_data['img_name'], f"check: {img_name} / {det_data['img_name']}"
    #             joint_img_h36m = det_data['pose2d'][:, :2]
    #             for i in range(self.human36_joint_num):
    #                 joint_img_h36m[i, :2] = affine_transform(joint_img_h36m[i, :2].copy(), trans)
    #             return joint_img_h36m

    # def compute_joint_err(self, pred_joint, target_joint):
    #     # root align joint
    #     pred_joint, target_joint = pred_joint - pred_joint[:, :1, :], target_joint - target_joint[:, :1, :]

    #     pred_joint, target_joint = pred_joint.detach().cpu().numpy(), target_joint.detach().cpu().numpy()

    #     pred_joint, target_joint = pred_joint[:, self.human36_eval_joint, :], target_joint[:, self.human36_eval_joint, :]
    #     joint_mean_error = np.power((np.power((pred_joint - target_joint), 2)).sum(axis=2), 0.5).mean()

    #     return joint_mean_error

    # def compute_both_err(self, pred_mesh, target_mesh, pred_joint, target_joint):
    #     # root align joint
    #     pred_mesh, target_mesh = pred_mesh - pred_joint[:, :1, :], target_mesh - target_joint[:, :1, :]
    #     pred_joint, target_joint = pred_joint - pred_joint[:, :1, :], target_joint - target_joint[:, :1, :]

    #     pred_mesh, target_mesh = pred_mesh.detach().cpu().numpy(), target_mesh.detach().cpu().numpy()
    #     pred_joint, target_joint = pred_joint.detach().cpu().numpy(), target_joint.detach().cpu().numpy()

    #     pred_joint, target_joint = pred_joint[:, self.human36_eval_joint, :], target_joint[:, self.human36_eval_joint, :]
    #     mesh_mean_error = np.power((np.power((pred_mesh - target_mesh), 2)).sum(axis=2), 0.5).mean()
    #     joint_mean_error = np.power((np.power((pred_joint - target_joint), 2)).sum(axis=2), 0.5).mean()

    #     return joint_mean_error, mesh_mean_error

    # def evaluate_joint(self, outs):
    #     print('Evaluation start...')
    #     annots = self.datalist
    #     assert len(annots) == len(outs)
    #     sample_num = len(annots)

    #     mpjpe = np.zeros((sample_num, len(self.human36_eval_joint)))
    #     pampjpe = np.zeros((sample_num, len(self.human36_eval_joint)))
    #     for n in range(sample_num):
    #         out = outs[n]
    #         annot = annots[n]

    #         # render materials
    #         pose_coord_out, pose_coord_gt = out['joint_coord'], annot['joint_cam']

    #         # root joint alignment
    #         pose_coord_out, pose_coord_gt = pose_coord_out - pose_coord_out[:1], pose_coord_gt - pose_coord_gt[:1]
    #         # sample eval joitns
    #         pose_coord_out, pose_coord_gt = pose_coord_out[self.human36_eval_joint, :], pose_coord_gt[self.human36_eval_joint, :]

    #         # pose error calculate
    #         mpjpe[n] = np.sqrt(np.sum((pose_coord_out - pose_coord_gt) ** 2, 1))
    #         # perform rigid alignment
    #         pose_coord_out = rigid_align(pose_coord_out, pose_coord_gt)
    #         pampjpe[n] = np.sqrt(np.sum((pose_coord_out - pose_coord_gt) ** 2, 1))

    #     # total pose error
    #     tot_err = np.mean(mpjpe)
    #     eval_summary = 'MPJPE (mm)    >> tot: %.2f\n' % (tot_err)
    #     print(eval_summary)

    #     tot_err = np.mean(pampjpe)
    #     eval_summary = 'PA-MPJPE (mm) >> tot: %.2f\n' % (tot_err)
    #     print(eval_summary)

    # def evaluate(self, outs):
    #     print('Evaluation start...')
    #     annots = self.datalist
    #     assert len(annots) == len(outs)
    #     sample_num = len(outs)

    #     # eval H36M joints
    #     pose_error_h36m = np.zeros((sample_num, len(self.human36_eval_joint)))  # pose error
    #     pose_error_action_h36m = [[] for _ in range(len(self.action_name))]  # pose error for each sequence
    #     pose_pa_error_h36m = np.zeros((sample_num, len(self.human36_eval_joint)))  # pose error
    #     pose_pa_error_action_h36m = [[] for _ in range(len(self.action_name))]  # pose error for each sequence

    #     # eval SMPL joints and mesh vertices
    #     pose_error = np.zeros((sample_num, self.smpl_joint_num))  # pose error
    #     pose_error_action = [[] for _ in range(len(self.action_name))]  # pose error for each sequence
    #     mesh_error = np.zeros((sample_num, self.smpl_vertex_num))  # mesh error
    #     mesh_error_action = [[] for _ in range(len(self.action_name))]  # mesh error for each sequence
    #     for n in range(sample_num):
    #         annot = annots[n]
    #         out = outs[n]

    #         # render materials
    #         img_path = annot['img_path']
    #         obj_name = '_'.join(img_path.split('/')[-2:])[:-4]

    #         # root joint alignment
    #         mesh_coord_out, mesh_coord_gt = out['mesh_coord'], out['mesh_coord_target']
    #         joint_coord_out, joint_coord_gt = np.dot(self.joint_regressor_smpl, mesh_coord_out), np.dot(self.joint_regressor_smpl, mesh_coord_gt)
    #         mesh_coord_out = mesh_coord_out - joint_coord_out[self.smpl_root_joint_idx:self.smpl_root_joint_idx+1]
    #         mesh_coord_gt = mesh_coord_gt - joint_coord_gt[self.smpl_root_joint_idx:self.smpl_root_joint_idx+1]
    #         pose_coord_out = joint_coord_out - joint_coord_out[self.smpl_root_joint_idx:self.smpl_root_joint_idx+1]
    #         pose_coord_gt = joint_coord_gt - joint_coord_gt[self.smpl_root_joint_idx:self.smpl_root_joint_idx+1]

    #         # pose error calculate
    #         pose_error[n] = np.sqrt(np.sum((pose_coord_out - pose_coord_gt) ** 2, 1))
    #         img_name = annot['img_path']
    #         action_idx = int(img_name[img_name.find('act') + 4:img_name.find('act') + 6]) - 2
    #         pose_error_action[action_idx].append(pose_error[n].copy())

    #         # mesh error calculate
    #         mesh_error[n] = np.sqrt(np.sum((mesh_coord_out - mesh_coord_gt) ** 2, 1))
    #         img_name = annot['img_path']
    #         action_idx = int(img_name[img_name.find('act') + 4:img_name.find('act') + 6]) - 2
    #         mesh_error_action[action_idx].append(mesh_error[n].copy())

    #         # pose error of h36m calculate
    #         pose_coord_out_h36m = np.dot(self.joint_regressor_human36, mesh_coord_out)
    #         pose_coord_out_h36m = pose_coord_out_h36m - pose_coord_out_h36m[self.human36_root_joint_idx]
    #         pose_coord_out_h36m = pose_coord_out_h36m[self.human36_eval_joint, :]
    #         pose_coord_gt_h36m = annot['joint_cam']
    #         pose_coord_gt_h36m = pose_coord_gt_h36m - pose_coord_gt_h36m[self.human36_root_joint_idx]
    #         pose_coord_gt_h36m = pose_coord_gt_h36m[self.human36_eval_joint, :]
    #         pose_error_h36m[n] = np.sqrt(np.sum((pose_coord_out_h36m - pose_coord_gt_h36m) ** 2, 1))
    #         pose_coord_out_h36m = rigid_align(pose_coord_out_h36m, pose_coord_gt_h36m) # perform rigid alignment
    #         pose_pa_error_h36m[n] = np.sqrt(np.sum((pose_coord_out_h36m - pose_coord_gt_h36m) ** 2, 1))
    #         img_name = annot['img_path']
    #         action_idx = int(img_name[img_name.find('act') + 4:img_name.find('act') + 6]) - 2
    #         pose_error_action_h36m[action_idx].append(pose_error_h36m[n].copy())
    #         pose_pa_error_action_h36m[action_idx].append(pose_pa_error_h36m[n].copy())

    #         vis = cfg.TEST.vis
    #         if vis and (n % 500 == 0):
    #             mesh_to_save = mesh_coord_out / 1000
    #             obj_path = osp.join(cfg.vis_dir, f'{obj_name}.obj')
    #             save_obj(mesh_to_save, self.mesh_model.face, obj_path)

    #     # total pose error (H36M joint set)
    #     tot_err = np.mean(pose_error_h36m)
    #     metric = 'MPJPE'
    #     eval_summary = 'Protocol ' + str(self.protocol) + ' H36M pose error (' + metric + ') >> tot: %.2f\n' % (tot_err)
    #     # pose error for each action
    #     for i in range(len(pose_error_action_h36m)):
    #         err = np.mean(np.array(pose_error_action_h36m[i]))
    #         eval_summary += (self.action_name[i] + ': %.2f ' % err)
    #     print(eval_summary)

    #     tot_err = np.mean(pose_pa_error_h36m)
    #     metric = 'PA-MPJPE'
    #     eval_summary = 'Protocol ' + str(self.protocol) + ' H36M pose error (' + metric + ') >> tot: %.2f\n' % (tot_err)
    #     # pose error for each action
    #     for i in range(len(pose_pa_error_action_h36m)):
    #         err = np.mean(np.array(pose_pa_error_action_h36m[i]))
    #         eval_summary += (self.action_name[i] + ': %.2f ' % err)
    #     print(eval_summary)

    #     # total pose error (SMPL joint set)
    #     tot_err = np.mean(pose_error)
    #     metric = 'MPJPE'
    #     eval_summary = 'Protocol ' + str(self.protocol) + ' SMPL pose error (' + metric + ') >> tot: %.2f\n' % (tot_err)
    #     # pose error for each action
    #     for i in range(len(pose_error_action)):
    #         err = np.mean(np.array(pose_error_action[i]))
    #         eval_summary += (self.action_name[i] + ': %.2f ' % err)
    #     print(eval_summary)

    #     # total mesh error
    #     tot_err = np.mean(mesh_error)
    #     metric = 'MPVPE'
    #     eval_summary = 'Protocol ' + str(self.protocol) + ' SMPL mesh error (' + metric + ') >> tot: %.2f\n' % (tot_err)
    #     # mesh error for each action
    #     for i in range(len(mesh_error_action)):
    #         err = np.mean(np.array(mesh_error_action[i]))
    #         eval_summary += (self.action_name[i] + ': %.2f ' % err)
    #     print(eval_summary)



"""time and date"""
import time


def get_time_str(time_float: float = None) -> str:
    """获取当前时间字符串

    Args:
        time_float (float, optional): time.time(). Defaults to None.

    Returns:
        str: 日期时间, 格式: 2021.05.17 12:20:20
    """

    if time_float is None:
        time_float = time.time()
    time_local = time.localtime(time_float)
    return time.strftime("%Y.%m.%d %H:%M:%S", time_local)


def get_date_str(time_float: float = None) -> str:
    """获取日期

    Args:
        time_float (float, optional): time.time(). Defaults to None.

    Returns:
        str: 格式: 2021.05.17
    """
    return get_time_str(time_float).split(" ")[0]


def get_time_folder() -> str:
    """获取以时间命名的文件夹名称

    Returns:
        str: 格式: 20210517_122020
    """

    return time.strftime("%Y%m%d_%H%M%S", time.localtime(time.time()))



"""downloader"""
from urllib.request import Request, urlopen

import requests

from .logger import Logger


def download_file(url: str, filename: str) -> bool:
    """download file by url

    Args:
        url (str): url
        filename (str): filename

    Returns:
        bool: 是否下载成功
    """
    try:
        content = requests.get(url).content
    except Exception as e:
        Logger.error(e)
        return False
    with open(filename, "wb") as fw:
        fw.write(content)
    return True


def download_web_text(url: str) -> str:
    """下载网页源码

    Args:
        url (str): url

    Returns:
        str: 网页源码
    """
    try:
        headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:23.0) Gecko/20100101 Firefox/23.0"}
        req = Request(url=url, headers=headers)
        content = urlopen(req).read().decode("utf-8")
        return content
    except Exception as e:
        print(e)
        return ""



"""同步数据
rsync 教程
https://www.ruanyifeng.com/blog/2020/08/rsync.html
"""
import os
import os.path as osp


def __fix_path(path: str) -> str:
    if not path.endswith("/"):
        path = path + "/"
    return path


def rsync(src_path: str, dst_path: str, delete: bool = False):
    """同步数据 copy src_path to dst_path

    Args:
        src_path (str): 源头文件夹
        dst_path (str): 目标文件夹
        delete (bool): 是否删除目标文件夹多余的文件(镜像)
    """
    if not osp.exists(dst_path):
        os.makedirs(dst_path, exist_ok=True)

    src_path = __fix_path(src_path)
    dst_path = __fix_path(dst_path)
    delete_text = " --delete " if delete else " "
    order = f"rsync -atvz{delete_text}{src_path} {dst_path}"
    print("[order]", order)
    os.system(order)


def rsync_ssh_upload(local_path: str, server_path: str, ip: str, port: int, user: str, password: str, delete: bool = False):
    """上传数据

    Args:
        local_path (str): 本地文件夹
        server_path (str): 服务器文件夹
        ip (str): 服务器IP
        port (int): 端口
        user (str): 用户名
        password (str): 密码
        delete (bool): 是否删除目标文件夹多余的文件(镜像)
    """
    local_path = __fix_path(local_path)
    server_path = __fix_path(server_path)
    delete_text = " --delete " if delete else " "
    order = f"sshpass -p {password} rsync -atvz{delete_text}-e 'ssh -p {port}' {local_path} {user}@{ip}:{server_path}"
    print("[order]", order)
    os.system(order)


def rsync_ssh_download(server_path: str, local_path: str, ip: str, port: int, user: str, password: str, delete: bool = False):
    """下载数据

    Args:
        server_path (str): 服务器文件夹
        local_path (str): 本地文件夹
        ip (str): 服务器IP
        port (int): 端口
        user (str): 用户名
        password (str): 密码
        delete (bool): 是否删除目标文件夹多余的文件(镜像)
    """
    if not osp.exists(local_path):
        os.makedirs(local_path, exist_ok=True)
    server_path = __fix_path(server_path)
    local_path = __fix_path(local_path)
    delete_text = " --delete " if delete else " "
    order = f"sshpass -p {password} rsync -atvz{delete_text} -e 'ssh -p {port}' {user}@{ip}:{server_path} {local_path}"
    print("[order]", order)
    os.system(order)



"""多进程处理任务"""
import time
from multiprocessing import Manager, Process
from typing import List

from tqdm import tqdm


def multi_tasks(func, data_list: List, pbar_name: str = "waiting", workers: int = 10, refresh_time: float = 0.1) -> None:
    """_summary_

    Args:
        func (_type_): 处理函数
        data_list (List): 需要处理的数据列表
        pbar_name (str, optional): 进度条显示名称. Defaults to "waiting".
        workers (int, optional): 进程数量. Defaults to 10.
        refresh_time (float, optional): 进度条更新时间(秒). Defaults to 0.1.
    """
    results = Manager().list()
    process_list: List[Process] = []

    step = len(data_list) // workers + 1
    all_idx = list(range(len(data_list)))
    for i in range(0, len(data_list), step):
        part_idxs = all_idx[i : i + step]
        process_list.append(Process(target=_batch_func, args=(func, data_list, part_idxs, results)))
    process_list.append(Process(target=_show_processon, args=(len(data_list), pbar_name, refresh_time, results)))
    for p in process_list:
        p.start()
    for p in process_list:
        p.join()

    real_results = [None] * len(data_list)
    for i, result in results:
        real_results[i] = result
    return real_results


def _show_processon(nums: int, pbar_name: str, refresh_time: float, results):
    """显示进度条"""
    cnt = 0
    step = 0
    pbar = tqdm(range(nums), desc=pbar_name)
    while cnt < nums:
        cnt = len(results)
        pbar.update(cnt - step)
        step = cnt
        time.sleep(refresh_time)


def _batch_func(func, data_list: List, part_idxs: List[int], results):
    n = func.__code__.co_argcount
    for idx in part_idxs:
        data = data_list[idx]
        if n == 1:
            result = func(data)
        else:
            result = func(*data)
        results.append([idx, result])


def test_tasks():
    """测试"""
    multi_tasks(lambda x: print(x), [1, 2, 3, 4, 5], pbar_name="print")



import logging
import os
import os.path as osp
import platform

import logzero
from logzero import logger

# python3.8版本之后才支持 stacklevel
version = platform.python_version()
version = [int(a) for a in version.split(".")]
has_stacklevel = not (version[0] <= 3 and version[1] < 8)


class Logger:
    """日志"""

    _logfilename = None

    @staticmethod
    def debug(msg: str):
        """debug

        Args:
            msg (str): 需要打印的字符串
        """
        if has_stacklevel:
            logger.debug(msg, stacklevel=2)
        else:
            logger.debug(msg)

    @staticmethod
    def info(msg: str):
        """info

        Args:
            msg (str): 需要打印的字符串
        """
        if has_stacklevel:
            logger.info(msg, stacklevel=2)
        else:
            logger.info(msg)

    @staticmethod
    def warn(msg: str):
        """warn

        Args:
            msg (str): 需要打印的字符串
        """
        if has_stacklevel:
            logger.warning(msg, stacklevel=2)
        else:
            logger.warning(msg)

    @staticmethod
    def error(msg):
        """error

        Args:
            msg (str): 需要打印的字符串
        """
        if has_stacklevel:
            logger.error(msg, stacklevel=2)
        else:
            logger.error(msg)

    @staticmethod
    def logfile(filename: str, clear: bool = False):
        """设置日志文件名

        Args:
            filename (str): 日志文件名
            clear (bool, optional): 是否清除日志文件. Defaults to False.
        """
        path = osp.dirname(filename)
        if path and (not osp.exists(path)):
            os.makedirs(path, exist_ok=True)

        msg = f"logfile -> {filename}"
        if has_stacklevel:
            logger.warning(msg, stacklevel=2)
        else:
            logger.warning(msg)
        if clear and osp.exists(filename):
            os.remove(filename)
        logzero.logfile(filename)
        Logger._logfilename = filename

    @staticmethod
    def loglevel(level):
        """设置日志等级: debug, info, warn, error"""
        level = level.upper()
        logzero.loglevel(eval(f"logging.{level}"))

        msg = f"loglevel -> {level}"
        if has_stacklevel:
            logger.warning(msg, stacklevel=2)
        else:
            logger.warning(msg)

    @staticmethod
    def print(*text_list, end="\n"):
        """print

        Args:
            end (str, optional): 分隔符. Defaults to "\n".
        """
        if len(text_list) == 0:
            text = "\n"
            end = ""
        else:
            text = " ".join([str(t) for t in text_list])

        if Logger._logfilename is not None:
            with open(Logger._logfilename, "a") as fw:
                fw.write(text + end)
        else:
            print(text, end=end)


Logger.loglevel("debug")



import sys

import cv2
import numpy as np
from PySide6 import QtCore, QtGui
from PySide6.QtGui import QScreen
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QLabel, QSlider,
                               QVBoxLayout, QWidget)
from torch.utils.data import Dataset

from ..datasets import ImageFolderDataset, VideoDataset
from .pose_3d_widget import Pose3DWidget

__all__ = ["show_dataset"]


class ImageWidget(QLabel):
    """显示图片"""

    def __init__(self, img: np.ndarray = None):
        super().__init__()
        if img is not None:
            self.set_img(img)

    def set_img(self, img: np.ndarray):
        height, width, channel = img.shape
        assert channel == 3
        qImg = QtGui.QImage(img.data, width, height, width * channel, QtGui.QImage.Format_BGR888)
        self.setPixmap(QtGui.QPixmap.fromImage(qImg))

    def set_imgname(self, imgname):
        self.set_img(cv2.imread(imgname))


class DatasetWidget(QWidget):
    def __init__(self, dataset: Dataset = None) -> None:
        super().__init__()

        class Slider(QSlider):
            def __init__(self, max_num: int = 100):
                super().__init__(QtGui.Qt.Horizontal)
                self.set_max(max_num)

            def set_max(self, max_num: int):
                self.setMinimum(0)
                self.setMaximum(max_num)
                self.setSingleStep(1)
                self.setTickInterval(max(1, int(max_num / 10.0)))
                self.setTickPosition(QSlider.TicksBelow)

        layout1 = QVBoxLayout()
        layout = QHBoxLayout()
        self.slider = Slider()
        self.w_img = ImageWidget()
        self.w_pose_3d = Pose3DWidget()
        layout1.addWidget(self.slider)
        layout1.addWidget(self.w_img)
        layout.addLayout(layout1, 3)
        layout.addWidget(self.w_pose_3d, 1)
        self.setLayout(layout)
        self.slider.valueChanged.connect(self.valuechange)

        if dataset is not None:
            self.set_dataset(dataset)
            self.set_index(0)

    def set_dataset(self, dataset):
        self.dataset = dataset
        self.slider.set_max(len(self.dataset) - 1)

    def valuechange(self):
        self.set_index(self.slider.value())

    def set_index(self, index):
        d = self.dataset.render(index)
        if isinstance(d, dict):
            self.w_img.set_img(d["img"])
            if "pose_3d" in d:
                self.w_pose_3d.render(d["pose_3d"])
        else:
            self.w_img.set_img(d)

    def get_img(self, index):
        """获取数据集需要显示的内容"""
        if hasattr(self.dataset, "render"):
            return self.dataset.render(index)
        else:
            return self.dataset[index]["img"]


class DatasetViewer(QWidget):
    """数据集可视化界面"""

    def __init__(self, dataset: Dataset):
        super().__init__()

        self.setMouseTracking(True)

        layout = QVBoxLayout()
        layout.addWidget(DatasetWidget(dataset))
        self.setLayout(layout)
        # 软件图标
        # self.setWindowIcon(QtGui.QIcon(BOX_ICON["scholar"]))
        self.setWindowTitle("DatasetViewer")
        self.showMaximized()
        self.show()


def center(window):
    center = QScreen.availableGeometry(QApplication.primaryScreen()).center()
    geo = window.frameGeometry()
    geo.moveCenter(center)
    window.move(geo.topLeft())


def show_dataset(dataset):
    """数据集可视化界面"""
    app = QApplication()
    main_window = DatasetViewer(dataset)
    center(main_window)

    sys.exit(app.exec())


if __name__ == "__main__":
    show_dataset(VideoDataset("xxx.mp4"))



import numpy as np
from torch.utils.data import Dataset


class RandomSplitDataset(Dataset):
    """随机划分部分数据集"""

    def __init__(self, dataset: Dataset, random_split: float = 1.0) -> None:
        super().__init__()

        self.dataset = dataset
        self.random_split = random_split

        N = len(self.dataset)
        n = int(N * float(random_split))
        self.__index_list = np.random.permutation(N)[:n].tolist()

    def __len__(self):
        return len(self.__index_list)

    def __getitem__(self, index):
        real_index = self.__index_list[index]
        return self.dataset.__getitem__(real_index)



# -*- coding: utf-8 -*-

# Max-Planck-Gesellschaft zur Förderung der Wissenschaften e.V. (MPG) is
# holder of all proprietary rights on this computer program.
# You can only use this computer program if you have closed
# a license agreement with MPG or you get the right to use the computer
# program from someone who is authorized to grant you that right.
# Any use of the computer program without a valid license is prohibited and
# liable to prosecution.
#
# Copyright©2019 Max-Planck-Gesellschaft zur Förderung
# der Wissenschaften e.V. (MPG). acting on behalf of its Max Planck Institute
# for Intelligent Systems. All rights reserved.
#
# Contact: ps-license@tuebingen.mpg.de

# copy from https://github.com/mkocabas/PARE/blob/master/pare/utils/kp_utils.py

import numpy as np


def keypoint_hflip(kp, img_width):
    # Flip a keypoint horizontally around the y-axis
    # kp N,2
    if len(kp.shape) == 2:
        kp[:,0] = (img_width - 1.) - kp[:,0]
    elif len(kp.shape) == 3:
        kp[:, :, 0] = (img_width - 1.) - kp[:, :, 0]
    return kp


def convert_kps(joints2d, src, dst):
    src_names = eval(f'get_{src}_joint_names')()
    dst_names = eval(f'get_{dst}_joint_names')()

    out_joints2d = np.zeros((joints2d.shape[0], len(dst_names), joints2d.shape[-1]))

    for idx, jn in enumerate(dst_names):
        if jn in src_names:
            out_joints2d[:, idx] = joints2d[:, src_names.index(jn)]

    return out_joints2d


def get_perm_idxs(src, dst):
    src_names = eval(f'get_{src}_joint_names')()
    dst_names = eval(f'get_{dst}_joint_names')()
    idxs = [src_names.index(h) for h in dst_names if h in src_names]
    return idxs


def get_mpii3d_test_joint_names():
    return [
        'headtop', # 'head_top',
        'neck',
        'rshoulder',# 'right_shoulder',
        'relbow',# 'right_elbow',
        'rwrist',# 'right_wrist',
        'lshoulder',# 'left_shoulder',
        'lelbow', # 'left_elbow',
        'lwrist', # 'left_wrist',
        'rhip', # 'right_hip',
        'rknee', # 'right_knee',
        'rankle',# 'right_ankle',
        'lhip',# 'left_hip',
        'lknee',# 'left_knee',
        'lankle',# 'left_ankle'
        'hip',# 'pelvis',
        'Spine (H36M)',# 'spine',
        'Head (H36M)',# 'head'
    ]


def get_mpii3d_joint_names():
    return [
        'spine3', # 0,
        'spine4', # 1,
        'spine2', # 2,
        'Spine (H36M)', #'spine', # 3,
        'hip', # 'pelvis', # 4,
        'neck', # 5,
        'Head (H36M)', # 'head', # 6,
        "headtop", # 'head_top', # 7,
        'left_clavicle', # 8,
        "lshoulder", # 'left_shoulder', # 9,
        "lelbow", # 'left_elbow',# 10,
        "lwrist", # 'left_wrist',# 11,
        'left_hand',# 12,
        'right_clavicle',# 13,
        'rshoulder',# 'right_shoulder',# 14,
        'relbow',# 'right_elbow',# 15,
        'rwrist',# 'right_wrist',# 16,
        'right_hand',# 17,
        'lhip', # left_hip',# 18,
        'lknee', # 'left_knee',# 19,
        'lankle', #left ankle # 20
        'left_foot', # 21
        'left_toe', # 22
        "rhip", # 'right_hip',# 23
        "rknee", # 'right_knee',# 24
        "rankle", #'right_ankle', # 25
        'right_foot',# 26
        'right_toe' # 27
    ]


# def get_insta_joint_names():
#     return [
#         'rheel'            ,   # 0
#         'rknee'            ,   # 1
#         'rhip'             ,   # 2
#         'lhip'             ,   # 3
#         'lknee'            ,   # 4
#         'lheel'            ,   # 5
#         'rwrist'           ,   # 6
#         'relbow'           ,   # 7
#         'rshoulder'        ,   # 8
#         'lshoulder'        ,   # 9
#         'lelbow'           ,   # 10
#         'lwrist'           ,   # 11
#         'neck'             ,   # 12
#         'headtop'          ,   # 13
#         'nose'             ,   # 14
#         'leye'             ,   # 15
#         'reye'             ,   # 16
#         'lear'             ,   # 17
#         'rear'             ,   # 18
#         'lbigtoe'          ,   # 19
#         'rbigtoe'          ,   # 20
#         'lsmalltoe'        ,   # 21
#         'rsmalltoe'        ,   # 22
#         'lankle'           ,   # 23
#         'rankle'           ,   # 24
#     ]


def get_insta_joint_names():
    return [
        'OP RHeel',
        'OP RKnee',
        'OP RHip',
        'OP LHip',
        'OP LKnee',
        'OP LHeel',
        'OP RWrist',
        'OP RElbow',
        'OP RShoulder',
        'OP LShoulder',
        'OP LElbow',
        'OP LWrist',
        'OP Neck',
        'headtop',
        'OP Nose',
        'OP LEye',
        'OP REye',
        'OP LEar',
        'OP REar',
        'OP LBigToe',
        'OP RBigToe',
        'OP LSmallToe',
        'OP RSmallToe',
        'OP LAnkle',
        'OP RAnkle',
    ]


def get_mmpose_joint_names():
    # this naming is for the first 23 joints of MMPose
    # does not include hands and face
    return [
        'OP Nose', # 1
        'OP LEye', # 2
        'OP REye', # 3
        'OP LEar', # 4
        'OP REar', # 5
        'OP LShoulder', # 6
        'OP RShoulder', # 7
        'OP LElbow', # 8
        'OP RElbow', # 9
        'OP LWrist', # 10
        'OP RWrist', # 11
        'OP LHip', # 12
        'OP RHip', # 13
        'OP LKnee', # 14
        'OP RKnee', # 15
        'OP LAnkle', # 16
        'OP RAnkle', # 17
        'OP LBigToe', # 18
        'OP LSmallToe', # 19
        'OP LHeel', # 20
        'OP RBigToe', # 21
        'OP RSmallToe', # 22
        'OP RHeel', # 23
    ]


def get_insta_skeleton():
    return np.array(
        [
            [0 , 1],
            [1 , 2],
            [2 , 3],
            [3 , 4],
            [4 , 5],
            [6 , 7],
            [7 , 8],
            [8 , 9],
            [9 ,10],
            [2 , 8],
            [3 , 9],
            [10,11],
            [8 ,12],
            [9 ,12],
            [12,13],
            [12,14],
            [14,15],
            [14,16],
            [15,17],
            [16,18],
            [0 ,20],
            [20,22],
            [5 ,19],
            [19,21],
            [5 ,23],
            [0 ,24],
        ])


def get_staf_skeleton():
    return np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 4],
            [1, 5],
            [5, 6],
            [6, 7],
            [1, 8],
            [8, 9],
            [9, 10],
            [10, 11],
            [8, 12],
            [12, 13],
            [13, 14],
            [0, 15],
            [0, 16],
            [15, 17],
            [16, 18],
            [2, 9],
            [5, 12],
            [1, 19],
            [20, 19],
        ]
    )


def get_staf_joint_names():
    return [
        'OP Nose', # 0,
        'OP Neck', # 1,
        'OP RShoulder', # 2,
        'OP RElbow', # 3,
        'OP RWrist', # 4,
        'OP LShoulder', # 5,
        'OP LElbow', # 6,
        'OP LWrist', # 7,
        'OP MidHip', # 8,
        'OP RHip', # 9,
        'OP RKnee', # 10,
        'OP RAnkle', # 11,
        'OP LHip', # 12,
        'OP LKnee', # 13,
        'OP LAnkle', # 14,
        'OP REye', # 15,
        'OP LEye', # 16,
        'OP REar', # 17,
        'OP LEar', # 18,
        'Neck (LSP)', # 19,
        'Top of Head (LSP)', # 20,
    ]


def get_spin_op_joint_names():
    return [
        'OP Nose',        # 0
        'OP Neck',        # 1
        'OP RShoulder',   # 2
        'OP RElbow',      # 3
        'OP RWrist',      # 4
        'OP LShoulder',   # 5
        'OP LElbow',      # 6
        'OP LWrist',      # 7
        'OP MidHip',      # 8
        'OP RHip',        # 9
        'OP RKnee',       # 10
        'OP RAnkle',      # 11
        'OP LHip',        # 12
        'OP LKnee',       # 13
        'OP LAnkle',      # 14
        'OP REye',        # 15
        'OP LEye',        # 16
        'OP REar',        # 17
        'OP LEar',        # 18
        'OP LBigToe',     # 19
        'OP LSmallToe',   # 20
        'OP LHeel',       # 21
        'OP RBigToe',     # 22
        'OP RSmallToe',   # 23
        'OP RHeel',       # 24
    ]


def get_openpose_joint_names():
    return [
        'OP Nose',        # 0
        'OP Neck',        # 1
        'OP RShoulder',   # 2
        'OP RElbow',      # 3
        'OP RWrist',      # 4
        'OP LShoulder',   # 5
        'OP LElbow',      # 6
        'OP LWrist',      # 7
        'OP MidHip',      # 8
        'OP RHip',        # 9
        'OP RKnee',       # 10
        'OP RAnkle',      # 11
        'OP LHip',        # 12
        'OP LKnee',       # 13
        'OP LAnkle',      # 14
        'OP REye',        # 15
        'OP LEye',        # 16
        'OP REar',        # 17
        'OP LEar',        # 18
        'OP LBigToe',     # 19
        'OP LSmallToe',   # 20
        'OP LHeel',       # 21
        'OP RBigToe',     # 22
        'OP RSmallToe',   # 23
        'OP RHeel',       # 24
    ]


def get_spin_joint_names():
    return [
        'OP Nose',        # 0
        'OP Neck',        # 1
        'OP RShoulder',   # 2
        'OP RElbow',      # 3
        'OP RWrist',      # 4
        'OP LShoulder',   # 5
        'OP LElbow',      # 6
        'OP LWrist',      # 7
        'OP MidHip',      # 8
        'OP RHip',        # 9
        'OP RKnee',       # 10
        'OP RAnkle',      # 11
        'OP LHip',        # 12
        'OP LKnee',       # 13
        'OP LAnkle',      # 14
        'OP REye',        # 15
        'OP LEye',        # 16
        'OP REar',        # 17
        'OP LEar',        # 18
        'OP LBigToe',     # 19
        'OP LSmallToe',   # 20
        'OP LHeel',       # 21
        'OP RBigToe',     # 22
        'OP RSmallToe',   # 23
        'OP RHeel',       # 24
        'rankle',         # 25
        'rknee',          # 26
        'rhip',           # 27
        'lhip',           # 28
        'lknee',          # 29
        'lankle',         # 30
        'rwrist',         # 31
        'relbow',         # 32
        'rshoulder',      # 33
        'lshoulder',      # 34
        'lelbow',         # 35
        'lwrist',         # 36
        'neck',           # 37
        'headtop',        # 38
        'hip',            # 39 'Pelvis (MPII)', # 39
        'thorax',         # 40 'Thorax (MPII)', # 40
        'Spine (H36M)',   # 41
        'Jaw (H36M)',     # 42
        'Head (H36M)',    # 43
        'nose',           # 44
        'leye',           # 45 'Left Eye', # 45
        'reye',           # 46 'Right Eye', # 46
        'lear',           # 47 'Left Ear', # 47
        'rear',           # 48 'Right Ear', # 48
    ]

def get_muco3dhp_joint_names():
    return [
        'headtop',
        'thorax',
        'rshoulder',
        'relbow',
        'rwrist',
        'lshoulder',
        'lelbow',
        'lwrist',
        'rhip',
        'rknee',
        'rankle',
        'lhip',
        'lknee',
        'lankle',
        'hip',
        'Spine (H36M)',
        'Head (H36M)',
        'R_Hand',
        'L_Hand',
        'R_Toe',
        'L_Toe'
    ]

def get_h36m_joint_names():
    return [
        'hip',  # 0
        'lhip',  # 1
        'lknee',  # 2
        'lankle',  # 3
        'rhip',  # 4
        'rknee',  # 5
        'rankle',  # 6
        'Spine (H36M)',  # 7
        'neck',  # 8
        'Head (H36M)',  # 9
        'headtop',  # 10
        'lshoulder',  # 11
        'lelbow',  # 12
        'lwrist',  # 13
        'rshoulder',  # 14
        'relbow',  # 15
        'rwrist',  # 16
    ]


def get_spin_skeleton():
    return np.array(
        [
            [0 , 1],
            [1 , 2],
            [2 , 3],
            [3 , 4],
            [1 , 5],
            [5 , 6],
            [6 , 7],
            [1 , 8],
            [8 , 9],
            [9 ,10],
            [10,11],
            [8 ,12],
            [12,13],
            [13,14],
            [0 ,15],
            [0 ,16],
            [15,17],
            [16,18],
            [21,19],
            [19,20],
            [14,21],
            [11,24],
            [24,22],
            [22,23],
            [0 ,38],
        ]
    )


def get_openpose_skeleton():
    return np.array(
        [
            [0 , 1],
            [1 , 2],
            [2 , 3],
            [3 , 4],
            [1 , 5],
            [5 , 6],
            [6 , 7],
            [1 , 8],
            [8 , 9],
            [9 ,10],
            [10,11],
            [8 ,12],
            [12,13],
            [13,14],
            [0 ,15],
            [0 ,16],
            [15,17],
            [16,18],
            [21,19],
            [19,20],
            [14,21],
            [11,24],
            [24,22],
            [22,23],
        ]
    )


def get_posetrack_joint_names():
    return [
        "nose",
        "neck",
        "headtop",
        "lear",
        "rear",
        "lshoulder",
        "rshoulder",
        "lelbow",
        "relbow",
        "lwrist",
        "rwrist",
        "lhip",
        "rhip",
        "lknee",
        "rknee",
        "lankle",
        "rankle"
    ]


def get_posetrack_original_kp_names():
    return [
        'nose',
        'head_bottom',
        'head_top',
        'left_ear',
        'right_ear',
        'left_shoulder',
        'right_shoulder',
        'left_elbow',
        'right_elbow',
        'left_wrist',
        'right_wrist',
        'left_hip',
        'right_hip',
        'left_knee',
        'right_knee',
        'left_ankle',
        'right_ankle'
    ]


def get_pennaction_joint_names():
   return [
       "headtop",   # 0
       "lshoulder", # 1
       "rshoulder", # 2
       "lelbow",    # 3
       "relbow",    # 4
       "lwrist",    # 5
       "rwrist",    # 6
       "lhip" ,     # 7
       "rhip" ,     # 8
       "lknee",     # 9
       "rknee" ,    # 10
       "lankle",    # 11
       "rankle"     # 12
   ]


def get_common_joint_names():
    return [
        "rankle",    # 0  "lankle",    # 0
        "rknee",     # 1  "lknee",     # 1
        "rhip",      # 2  "lhip",      # 2
        "lhip",      # 3  "rhip",      # 3
        "lknee",     # 4  "rknee",     # 4
        "lankle",    # 5  "rankle",    # 5
        "rwrist",    # 6  "lwrist",    # 6
        "relbow",    # 7  "lelbow",    # 7
        "rshoulder", # 8  "lshoulder", # 8
        "lshoulder", # 9  "rshoulder", # 9
        "lelbow",    # 10  "relbow",    # 10
        "lwrist",    # 11  "rwrist",    # 11
        "neck",      # 12  "neck",      # 12
        "headtop",   # 13  "headtop",   # 13
    ]


def get_common_paper_joint_names():
    return [
        "Right Ankle",    # 0  "lankle",    # 0
        "Right Knee",     # 1  "lknee",     # 1
        "Right Hip",      # 2  "lhip",      # 2
        "Left Hip",      # 3  "rhip",      # 3
        "Left Knee",     # 4  "rknee",     # 4
        "Left Ankle",    # 5  "rankle",    # 5
        "Right Wrist",    # 6  "lwrist",    # 6
        "Right Elbow",    # 7  "lelbow",    # 7
        "Right Shoulder", # 8  "lshoulder", # 8
        "Left Shoulder", # 9  "rshoulder", # 9
        "Left Elbow",    # 10  "relbow",    # 10
        "Left Wrist",    # 11  "rwrist",    # 11
        "Neck",      # 12  "neck",      # 12
        "Head",   # 13  "headtop",   # 13
    ]


def get_common_skeleton():
    return np.array(
        [
            [ 0, 1 ],
            [ 1, 2 ],
            [ 3, 4 ],
            [ 4, 5 ],
            [ 6, 7 ],
            [ 7, 8 ],
            [ 8, 2 ],
            [ 8, 9 ],
            [ 9, 3 ],
            [ 2, 3 ],
            [ 8, 12],
            [ 9, 10],
            [12, 9 ],
            [10, 11],
            [12, 13],
        ]
    )


def get_coco_joint_names():
    return [
        "nose",      # 0
        "leye",      # 1
        "reye",      # 2
        "lear",      # 3
        "rear",      # 4
        "lshoulder", # 5
        "rshoulder", # 6
        "lelbow",    # 7
        "relbow",    # 8
        "lwrist",    # 9
        "rwrist",    # 10
        "lhip",      # 11
        "rhip",      # 12
        "lknee",     # 13
        "rknee",     # 14
        "lankle",    # 15
        "rankle",    # 16
    ]


def get_ochuman_joint_names():
    return [
        'rshoulder',
        'relbow',
        'rwrist',
        'lshoulder',
        'lelbow',
        'lwrist',
        'rhip',
        'rknee',
        'rankle',
        'lhip',
        'lknee',
        'lankle',
        'headtop',
        'neck',
        'rear',
        'lear',
        'nose',
        'reye',
        'leye'
    ]


def get_crowdpose_joint_names():
    return [
        'lshoulder',
        'rshoulder',
        'lelbow',
        'relbow',
        'lwrist',
        'rwrist',
        'lhip',
        'rhip',
        'lknee',
        'rknee',
        'lankle',
        'rankle',
        'headtop',
        'neck'
    ]

def get_coco_skeleton():
    # 0  - nose,
    # 1  - leye,
    # 2  - reye,
    # 3  - lear,
    # 4  - rear,
    # 5  - lshoulder,
    # 6  - rshoulder,
    # 7  - lelbow,
    # 8  - relbow,
    # 9  - lwrist,
    # 10 - rwrist,
    # 11 - lhip,
    # 12 - rhip,
    # 13 - lknee,
    # 14 - rknee,
    # 15 - lankle,
    # 16 - rankle,
    return np.array(
        [
            [15, 13],
            [13, 11],
            [16, 14],
            [14, 12],
            [11, 12],
            [ 5, 11],
            [ 6, 12],
            [ 5, 6 ],
            [ 5, 7 ],
            [ 6, 8 ],
            [ 7, 9 ],
            [ 8, 10],
            [ 1, 2 ],
            [ 0, 1 ],
            [ 0, 2 ],
            [ 1, 3 ],
            [ 2, 4 ],
            [ 3, 5 ],
            [ 4, 6 ]
        ]
    )


def get_mpii_joint_names():
    return [
        "rankle",    # 0
        "rknee",     # 1
        "rhip",      # 2
        "lhip",      # 3
        "lknee",     # 4
        "lankle",    # 5
        "hip",       # 6
        "thorax",    # 7
        "neck",      # 8
        "headtop",   # 9
        "rwrist",    # 10
        "relbow",    # 11
        "rshoulder", # 12
        "lshoulder", # 13
        "lelbow",    # 14
        "lwrist",    # 15
    ]


def get_mpii_skeleton():
    # 0  - rankle,
    # 1  - rknee,
    # 2  - rhip,
    # 3  - lhip,
    # 4  - lknee,
    # 5  - lankle,
    # 6  - hip,
    # 7  - thorax,
    # 8  - neck,
    # 9  - headtop,
    # 10 - rwrist,
    # 11 - relbow,
    # 12 - rshoulder,
    # 13 - lshoulder,
    # 14 - lelbow,
    # 15 - lwrist,
    return np.array(
        [
            [ 0, 1 ],
            [ 1, 2 ],
            [ 2, 6 ],
            [ 6, 3 ],
            [ 3, 4 ],
            [ 4, 5 ],
            [ 6, 7 ],
            [ 7, 8 ],
            [ 8, 9 ],
            [ 7, 12],
            [12, 11],
            [11, 10],
            [ 7, 13],
            [13, 14],
            [14, 15]
        ]
    )


def get_aich_joint_names():
    return [
        "rshoulder", # 0
        "relbow",    # 1
        "rwrist",    # 2
        "lshoulder", # 3
        "lelbow",    # 4
        "lwrist",    # 5
        "rhip",      # 6
        "rknee",     # 7
        "rankle",    # 8
        "lhip",      # 9
        "lknee",     # 10
        "lankle",    # 11
        "headtop",   # 12
        "neck",      # 13
    ]


def get_aich_skeleton():
    # 0  - rshoulder,
    # 1  - relbow,
    # 2  - rwrist,
    # 3  - lshoulder,
    # 4  - lelbow,
    # 5  - lwrist,
    # 6  - rhip,
    # 7  - rknee,
    # 8  - rankle,
    # 9  - lhip,
    # 10 - lknee,
    # 11 - lankle,
    # 12 - headtop,
    # 13 - neck,
    return np.array(
        [
            [ 0, 1 ],
            [ 1, 2 ],
            [ 3, 4 ],
            [ 4, 5 ],
            [ 6, 7 ],
            [ 7, 8 ],
            [ 9, 10],
            [10, 11],
            [12, 13],
            [13, 0 ],
            [13, 3 ],
            [ 0, 6 ],
            [ 3, 9 ]
        ]
    )


def get_3dpw_joint_names():
    return [
        "nose",      # 0
        "thorax",    # 1
        "rshoulder", # 2
        "relbow",    # 3
        "rwrist",    # 4
        "lshoulder", # 5
        "lelbow",    # 6
        "lwrist",    # 7
        "rhip",      # 8
        "rknee",     # 9
        "rankle",    # 10
        "lhip",      # 11
        "lknee",     # 12
        "lankle",    # 13
    ]


def get_3dpw_skeleton():
    return np.array(
        [
            [ 0, 1 ],
            [ 1, 2 ],
            [ 2, 3 ],
            [ 3, 4 ],
            [ 1, 5 ],
            [ 5, 6 ],
            [ 6, 7 ],
            [ 2, 8 ],
            [ 5, 11],
            [ 8, 11],
            [ 8, 9 ],
            [ 9, 10],
            [11, 12],
            [12, 13]
        ]
    )


def get_smplcoco_joint_names():
    return [
        "rankle",    # 0
        "rknee",     # 1
        "rhip",      # 2
        "lhip",      # 3
        "lknee",     # 4
        "lankle",    # 5
        "rwrist",    # 6
        "relbow",    # 7
        "rshoulder", # 8
        "lshoulder", # 9
        "lelbow",    # 10
        "lwrist",    # 11
        "neck",      # 12
        "headtop",   # 13
        "nose",      # 14
        "leye",      # 15
        "reye",      # 16
        "lear",      # 17
        "rear",      # 18
    ]


def get_smplcoco_skeleton():
    return np.array(
        [
            [ 0, 1 ],
            [ 1, 2 ],
            [ 3, 4 ],
            [ 4, 5 ],
            [ 6, 7 ],
            [ 7, 8 ],
            [ 8, 12],
            [12, 9 ],
            [ 9, 10],
            [10, 11],
            [12, 13],
            [14, 15],
            [15, 17],
            [16, 18],
            [14, 16],
            [ 8, 2 ],
            [ 9, 3 ],
            [ 2, 3 ],
        ]
    )


def get_smpl_joint_names():
    return [
        'hips',            # 0
        'leftUpLeg',       # 1
        'rightUpLeg',      # 2
        'spine',           # 3
        'leftLeg',         # 4
        'rightLeg',        # 5
        'spine1',          # 6
        'leftFoot',        # 7
        'rightFoot',       # 8
        'spine2',          # 9
        'leftToeBase',     # 10
        'rightToeBase',    # 11
        'neck',            # 12
        'leftShoulder',    # 13
        'rightShoulder',   # 14
        'head',            # 15
        'leftArm',         # 16
        'rightArm',        # 17
        'leftForeArm',     # 18
        'rightForeArm',    # 19
        'leftHand',        # 20
        'rightHand',       # 21
        'leftHandIndex1',  # 22
        'rightHandIndex1', # 23
    ]


def get_smpl_paper_joint_names():
    return [
        'Hips',            # 0
        'Left Hip',       # 1
        'Right Hip',      # 2
        'Spine',           # 3
        'Left Knee',         # 4
        'Right Knee',        # 5
        'Spine_1',          # 6
        'Left Ankle',        # 7
        'Right Ankle',       # 8
        'Spine_2',          # 9
        'Left Toe',     # 10
        'Right Toe',    # 11
        'Neck',            # 12
        'Left Shoulder',    # 13
        'Right Shoulder',   # 14
        'Head',            # 15
        'Left Arm',         # 16
        'Right Arm',        # 17
        'Left Elbow',     # 18
        'Right Elbow',    # 19
        'Left Hand',        # 20
        'Right Hand',       # 21
        'Left Thumb',  # 22
        'Right Thumb', # 23
    ]


def get_smpl_neighbor_triplets():
    return [
        [ 0,  1, 2 ],  # 0
        [ 1,  4, 0 ],  # 1
        [ 2,  0, 5 ],  # 2
        [ 3,  0, 6 ],  # 3
        [ 4,  7, 1 ],  # 4
        [ 5,  2, 8 ],  # 5
        [ 6,  3, 9 ],  # 6
        [ 7, 10, 4 ],  # 7
        [ 8,  5, 11],  # 8
        [ 9, 13, 14],  # 9
        [10,  7, 4 ],  # 10
        [11,  8, 5 ],  # 11
        [12,  9, 15],  # 12
        [13, 16, 9 ],  # 13
        [14,  9, 17],  # 14
        [15,  9, 12],  # 15
        [16, 18, 13],  # 16
        [17, 14, 19],  # 17
        [18, 20, 16],  # 18
        [19, 17, 21],  # 19
        [20, 22, 18],  # 20
        [21, 19, 23],  # 21
        [22, 20, 18],  # 22
        [23, 19, 21],  # 23
    ]


def get_smpl_skeleton():
    return np.array(
        [
            [ 0, 1 ],
            [ 0, 2 ],
            [ 0, 3 ],
            [ 1, 4 ],
            [ 2, 5 ],
            [ 3, 6 ],
            [ 4, 7 ],
            [ 5, 8 ],
            [ 6, 9 ],
            [ 7, 10],
            [ 8, 11],
            [ 9, 12],
            [ 9, 13],
            [ 9, 14],
            [12, 15],
            [13, 16],
            [14, 17],
            [16, 18],
            [17, 19],
            [18, 20],
            [19, 21],
            [20, 22],
            [21, 23],
        ]
    )


def map_spin_joints_to_smpl():
    # this function primarily will be used to copy 2D keypoint
    # confidences to pose parameters
    return [
        [(39, 27, 28), 0],  # hip,lhip,rhip->hips
        [(28,), 1],  # lhip->leftUpLeg
        [(27,), 2],  # rhip->rightUpLeg
        [(41, 27, 28, 39), 3],  # Spine->spine
        [(29,), 4],  # lknee->leftLeg
        [(26,), 5],  # rknee->rightLeg
        [(41, 40, 33, 34,), 6],  # spine, thorax ->spine1
        [(30,), 7],  # lankle->leftFoot
        [(25,), 8],  # rankle->rightFoot
        [(40, 33, 34), 9],  # thorax,shoulders->spine2
        [(30,), 10],  # lankle -> leftToe
        [(25,), 11],  # rankle -> rightToe
        [(37, 42, 33, 34), 12],  # neck, shoulders -> neck
        [(34,), 13],  # lshoulder->leftShoulder
        [(33,), 14],  # rshoulder->rightShoulder
        [(33, 34, 38, 43, 44, 45, 46, 47, 48,), 15],  # nose, eyes, ears, headtop, shoulders->head
        [(34,), 16],  # lshoulder->leftArm
        [(33,), 17],  # rshoulder->rightArm
        [(35,), 18],  # lelbow->leftForeArm
        [(32,), 19],  # relbow->rightForeArm
        [(36,), 20],  # lwrist->leftHand
        [(31,), 21],  # rwrist->rightHand
        [(36,), 22],  # lhand -> leftHandIndex
        [(31,), 23],  # rhand -> rightHandIndex
    ]


def map_smpl_to_common():
    return [
        [(11, 8), 0], # rightToe, rightFoot -> rankle
        [(5,), 1], # rightleg -> rknee,
        [(2,), 2], # rhip
        [(1,), 3], # lhip
        [(4,), 4], # leftLeg -> lknee
        [(10, 7), 5], # lefttoe, leftfoot -> lankle
        [(21, 23), 6], # rwrist
        [(18,), 7], # relbow
        [(17, 14), 8],  # rshoulder
        [(16, 13), 9],  # lshoulder
        [(19,), 10],  # lelbow
        [(20, 22), 11],  # lwrist
        [(0, 3, 6, 9, 12), 12],  # neck
        [(15,), 13],  # headtop
    ]


def relation_among_spin_joints():
    # this function primarily will be used to copy 2D keypoint
    # confidences to 3D joints
    return [
        [(), 25],
        [(), 26],
        [(39,), 27],
        [(39,), 28],
        [(), 29],
        [(), 30],
        [(), 31],
        [(), 32],
        [(), 33],
        [(), 34],
        [(), 35],
        [(), 36],
        [(40,42,44,43,38,33,34,), 37],
        [(43,44,45,46,47,48,33,34,), 38],
        [(27,28,), 39],
        [(27,28,37,41,42,), 40],
        [(27,28,39,40,), 41],
        [(37,38,44,45,46,47,48,), 42],
        [(44,45,46,47,48,38,42,37,33,34,), 43],
        [(44,45,46,47,48,38,42,37,33,34), 44],
        [(44,45,46,47,48,38,42,37,33,34), 45],
        [(44,45,46,47,48,38,42,37,33,34), 46],
        [(44,45,46,47,48,38,42,37,33,34), 47],
        [(44,45,46,47,48,38,42,37,33,34), 48],
    ]


import torch
import numpy as np


def list2numpy(a: list):
    return np.array(a, dtype=np.float32)


def list2tensor(a: list):
    return torch.from_numpy(list2numpy(a))


def numpy2tensor(a: np.ndarray):
    return torch.from_numpy(a)



import sys
from PySide6.QtWidgets import QMainWindow, QApplication
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
import numpy as np


def axisEqual3D(ax):
    extents = np.array([getattr(ax, "get_{}lim".format(dim))() for dim in "xyz"])
    sz = extents[:, 1] - extents[:, 0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize / 2
    for ctr, dim in zip(centers, "xyz"):
        getattr(ax, "set_{}lim".format(dim))(ctr - r, ctr + r)


class Figure3D(object):
    def __init__(self, title:str="3D") -> None:
        self.fig = plt.figure(figsize=(10, 10))
        self.ax = self.fig.add_subplot(111, projection="3d")
        self.ax.set_box_aspect([1, 1, 1])
        # self.render(np.zeros([10, 3]))

        self.ax.set_xlabel("X axis")
        self.ax.set_ylabel("Y axis")
        self.ax.set_zlabel("Z axis")
        self.ax.set_title(title)
        # self.ax.legend()
        axisEqual3D(self.ax)

    def render(self, kps_3d, kps_line=None):
        # 绘制骨架
        self.ax.clear()
        kps_line = kps_line or []
        for a, b in kps_line:
            x = np.array([kps_3d[a, 0], kps_3d[b, 0]])
            y = np.array([kps_3d[a, 1], kps_3d[b, 1]])
            z = np.array([kps_3d[a, 2], kps_3d[b, 2]])
            self.ax.plot(x, z, -y, c="r", linewidth=1)

        for i in range(kps_3d.shape[0]):
            x = kps_3d[i, 0]
            y = kps_3d[i, 1]
            z = kps_3d[i, 2]
            self.ax.scatter(x, y, z, c="g", marker='o')


    def show(self):
        self.fig.show()
def vis_3d_pose(kps_3d, kps_line, title: str = "3D Pose"):
    fig = Figure3D(title)
    fig.render(kps_3d, kps_line)
    fig.show()


class Pose3DWidget(FigureCanvasQTAgg):
    def __init__(self, kps_line=None):

        self.figure3d = Figure3D()
        self.kps_line = kps_line
        # fig, ax = vis_3d_pose([], [], show=False)
        # self.ax = ax
        super().__init__(self.figure3d.fig)

    def render(self, pose_3d, kps_line=None):
        kps_line = kps_line or self.kps_line
        self.figure3d.render(pose_3d, kps_line)


# class MainWindow(QMainWindow):
#     def __init__(self, *args, **kwargs):
#         super(MainWindow, self).__init__(*args, **kwargs)

#         # Create the maptlotlib FigureCanvas object,
#         # which defines a single set of axes as self.axes.
#         w = Pose3DWidget(np.array([[1, 2, 3], [2, 3, 4], [1, 3, 4], [2, 2, 2]]), [[0, 1], [1, 2], [0, 2], [1, 3]])
#         self.setCentralWidget(w)

#         self.show()


# app = QApplication(sys.argv)
# w = MainWindow()
# app.exec_()
